/*
 * ffmpeg_stub.c
 * Copyright (C) 2010 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <windows.h>
#include "ffmpeg_stub.h"

#if defined (__FFMPEG__)
  #define AVUTIL_VER	"51"
  #define AVCODEC_VER	"54"
  #define AVFORMAT_VER	"54"
  #define SWSCALE_VER	"2"
#elif defined (__LIBAV__)
  #define AVUTIL_VER	"51"
  #define AVCODEC_VER	"54"
  #define AVFORMAT_VER	"54"
  #define SWSCALE_VER	"2"
#else
  #error "Define either __FFMPEG__ or __LIBAV__"
#endif

int ffmpeg_stub_valid=0;

static HINSTANCE hLibAvUtil=NULL;
static HINSTANCE hLibAvCodec=NULL;
static HINSTANCE hLibAvFormat=NULL;
static HINSTANCE hLibSwScale=NULL;

AVCodec *__mp3_decoder;
AVCodec *__ac3_decoder;
AVCodec *__flac_decoder;
AVCodec *__mpc7_decoder;

int ffmpeg_find_decoders()
{
  __ac3_decoder=__avcodec_find_decoder(CODEC_ID_AC3);
  __mp3_decoder=__avcodec_find_decoder(CODEC_ID_MP3);
  __flac_decoder=__avcodec_find_decoder(CODEC_ID_FLAC);
  __mpc7_decoder=__avcodec_find_decoder(CODEC_ID_MUSEPACK7);

  return 0;
}

#define DECL(sym) typeof(sym) *__##sym;
#include "symbols/avutil.h"
#include "symbols/avcodec.h"
#include "symbols/avformat.h"
#include "symbols/swscale.h"
#undef DECL

#define DECL(sym) \
  if (NULL==(__##sym=(void *)GetProcAddress(H_LIB, #sym))) { \
	error("Error loading \"" #sym "\" from \"" NAME "\"."); \
    goto fail; \
  }

int ffmpeg_stub_load(const char *path, char *p, void (*error)(char *))
{
#define H_LIB hLibAvUtil
#define NAME "avutil-" AVUTIL_VER ".dll"
  strcpy(p, "in_ffsox\\" NAME);

  if (NULL==(H_LIB=LoadLibrary(TEXT(path)))) {
	error("Error loading \"" NAME "\".");
    goto fail;
  }

#include "symbols/avutil.h"
#undef NAME
#undef H_LIB

#define H_LIB hLibAvCodec
#define NAME "avcodec-" AVCODEC_VER ".dll"
  strcpy(p, "in_ffsox\\" NAME);

  if (NULL==(H_LIB=LoadLibrary(TEXT(path)))) {
	error("Error loading \"" NAME "\".");
    goto fail;
  }

#include "symbols/avcodec.h"
#undef NAME
#undef H_LIB

#define H_LIB hLibAvFormat
#define NAME "avformat-" AVFORMAT_VER ".dll"
  strcpy(p, "in_ffsox\\" NAME);

  if (NULL==(H_LIB=LoadLibrary(TEXT(path)))) {
	error("Error loading \"" NAME "\".");
    goto fail;
  }

#include "symbols/avformat.h"
#undef NAME
#undef H_LIB

#define H_LIB hLibSwScale
#define NAME "swscale-" SWSCALE_VER ".dll"
  strcpy(p, "in_ffsox\\" NAME);

  if (NULL==(H_LIB=LoadLibrary(TEXT(path)))) {
	error("Error loading \"" NAME "\".");
    goto fail;
  }

#include "symbols/swscale.h"
#undef NAME
#undef H_LIB

  return ffmpeg_stub_valid=1;
fail:
  return ffmpeg_stub_unload();
}

#undef DECL

int ffmpeg_stub_unload()
{
  if (NULL!=hLibSwScale) {
    FreeLibrary(hLibSwScale);
    hLibSwScale=NULL;
  }

  if (NULL!=hLibAvFormat) {
    FreeLibrary(hLibAvFormat);
    hLibAvFormat=NULL;
  }

  if (NULL!=hLibAvCodec) {
    FreeLibrary(hLibAvCodec);
    hLibAvCodec=NULL;
  }

  if (NULL!=hLibAvUtil) {
    FreeLibrary(hLibAvUtil);
    hLibAvUtil=NULL;
  }

  return ffmpeg_stub_valid=0;
}
