/*
 * in_ffsox.c
 * Copyright (C) 2010-2012 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "plugin.h"
#include "vout.h"
#define __DYNAMIC__
#include "ffmpeg_stub.h"
#include "libsox_stub.h"
#include <libavutil/opt.h>

#define __FLOAT__
//#define __SYNC__

#if defined (__FFMPEG__)
  #define DECODER "FFmpeg"
#elif defined (__LIBAV__)
  #define DECODER "libav"
#endif

typedef struct stream {
  int id;
  AVStream *stream;
  AVCodecContext *dec;
  AVCodec *codec;
  double time_base;
} stream_t;

typedef struct audio {
  stream_t s;
  sox_effects_chain_t *chain;
#ifndef OLD
  AVFrame *frame;
#endif
  AVPacket p1;
  AVPacket p2;
//#ifdef OLD
  uint8_t *rp;
  uint8_t *mp;
//#endif
  int apply_gain;
  int start;
//#ifdef OLD
  int eof;
//#endif
  void (*settime)(int);
  int (*read_frame)(track_t *, AVPacket *);
  int done;
} audio_t;

typedef struct video {
  stream_t s;
  volatile int run;
  volatile int pause;
  HANDLE hEventWrite;
  packet_queue_t qa;
  packet_queue_t qv;

  int width;
  int height;
  int in_pix_fmt;
  int out_pix_fmt;

  vout_t *vout;
  int (*gettime)();
  double step;
  double pts;
  timer_t timer;
  AVFrame *frame;
  struct SwsContext *img_convert_ctx;

  AVPicture out_pict;
  AVPicture in_pict;

#ifdef __SYNC__
  struct {
    CRITICAL_SECTION cs;
    HANDLE hEvent;
    int count;
    int a;
    int v;
  } sync;
#endif // __SYNC__
} video_t;

struct track {
  AVFormatContext *ic;
  int64_t start_time;
  audio_t a;
  video_t v;
  samplebuf_t *buf;

  HANDLE hThreadSoX;
};

#ifdef OLD
typedef int (*decode_t)(AVCodecContext *, void *, int *, AVPacket *);
#endif

static track_t track_play;
static track_t track_info;
#ifdef OLD
DECLARE_ALIGNED(16,uint8_t,ff_samples[(AVCODEC_MAX_AUDIO_FRAME_SIZE*3)/2]);
#endif

#ifndef OLD
static void get_tags_md(AVDictionary *metadata, samplebuf_t *buf)
{
  AVDictionaryEntry *tag=NULL;

#define MAKE_INFO(id) \
  if (NULL!=(tag=__av_dict_get(metadata,#id,NULL,\
      AV_DICT_DONT_STRDUP_KEY|AV_DICT_DONT_STRDUP_VAL))) \
    _snprintf (buf->tag.id, sizeof buf->tag.id, "%s", tag->value);
#define MAKE_NEW_INFO(id) MAKE_INFO(id)
#define MAKE_ALTERNATE_INFO(id1, id2) \
  if (0==buf->tag.id1[0]&&NULL!=(tag=__av_dict_get(metadata,#id2,NULL,\
      AV_DICT_DONT_STRDUP_KEY|AV_DICT_DONT_STRDUP_VAL))) \
    _snprintf (buf->tag.id1, sizeof buf->tag.id1, "%s", tag->value);
#include "info.h"
#undef MAKE_ALTERNATE_INFO
#undef MAKE_NEW_INFO
#undef MAKE_INFO
}
#else // OLD
static void get_tags_md(AVMetadata *metadata, samplebuf_t *buf)
{
  AVMetadataTag *tag=NULL;

#define MAKE_INFO(id) \
  if (NULL!=(tag=__av_metadata_get(metadata,#id,NULL,\
      AV_METADATA_DONT_STRDUP_KEY|AV_METADATA_DONT_STRDUP_VAL))) \
    _snprintf (buf->tag.id, sizeof buf->tag.id, "%s", tag->value);
#define MAKE_NEW_INFO(id) MAKE_INFO(id)
#define MAKE_ALTERNATE_INFO(id1, id2) \
  if (0==buf->tag.id1[0]&&NULL!=(tag=__av_metadata_get(metadata,#id2,NULL,\
      AV_METADATA_DONT_STRDUP_KEY|AV_METADATA_DONT_STRDUP_VAL))) \
    _snprintf (buf->tag.id1, sizeof buf->tag.id1, "%s", tag->value);
#include "info.h"
#undef MAKE_ALTERNATE_INFO
#undef MAKE_NEW_INFO
#undef MAKE_INFO
}
#endif // OLD

static void in_ffsox_fill_format(track_t *track)
{
  samplebuf_t *buf=track->buf;
  video_t *v=&track->v;
  audio_t *a=&track->a;

  if (0!=a->s.dec->bit_rate) {
    if (NULL!=v->s.codec) {
      _snprintf(buf->formatinformation,
          sizeof buf->formatinformation,
          "Length: %d seconds"
          "\nDecoder: " DECODER
          "\n\nAudio Codec: %s"
          "\nBitrate: %d kb/s"
          "\nChannels: %d"
          "\nBits per sample (in): %d"
          "\nSample Rate (in): %.f Hz"
          "\nBits per sample (out): %d"
          "\nSample Rate (out): %.f Hz"
          "\n\nVideo Codec: %s"
          ,
          buf->length/1000,
          a->s.codec?a->s.codec->name:"undefined",
		  a->s.dec->bit_rate/1000,
          a->s.dec->channels,
          0<a->s.dec->bits_per_raw_sample
                  ?a->s.dec->bits_per_raw_sample
#ifndef OLD
                  :8*__av_get_bytes_per_sample(a->s.dec->sample_fmt),
#else
                  :__av_get_bits_per_sample_fmt(a->s.dec->sample_fmt),
#endif
          (double)a->s.dec->sample_rate,
          track->buf->bps,
          (double)track->buf->rate,
          v->s.codec?v->s.codec->name:"undefined");
    }
    else {
      _snprintf(buf->formatinformation,
          sizeof buf->formatinformation,
          "Length: %d seconds"
          "\nDecoder: " DECODER
          "\n\nAudio Codec: %s"
          "\nBitrate: %d kb/s"
          "\nChannels: %d"
          "\nBits per sample (in): %d"
          "\nSample Rate (in): %.f Hz"
          "\nBits per sample (out): %d"
          "\nSample Rate (out): %.f Hz"
          ,
          buf->length/1000,
          a->s.codec?a->s.codec->name:"undefinded",
		  a->s.dec->bit_rate/1000,
          a->s.dec->channels,
          0<a->s.dec->bits_per_raw_sample
                  ?a->s.dec->bits_per_raw_sample
#ifndef OLD
                  :8*__av_get_bytes_per_sample(a->s.dec->sample_fmt),
#else
                  :__av_get_bits_per_sample_fmt(a->s.dec->sample_fmt),
#endif
          (double)a->s.dec->sample_rate,
          track->buf->bps,
          (double)track->buf->rate);
    }
  }
  else {
    if (NULL!=v->s.codec) {
      _snprintf(buf->formatinformation,
          sizeof buf->formatinformation,
          "Length: %d seconds"
          "\nDecoder: " DECODER
          "\n\nAudio Codec: %s"
          "\nChannels: %d"
          "\nBits per sample (in): %d"
          "\nSample Rate (in): %.f Hz"
          "\nBits per sample (out): %d"
          "\nSample Rate (out): %.f Hz"
          "\n\nVideo Codec: %s"
          ,
          buf->length/1000,
          a->s.codec?a->s.codec->name:"undefined",
          a->s.dec->channels,
          0<a->s.dec->bits_per_raw_sample
                  ?a->s.dec->bits_per_raw_sample
#ifndef OLD
                  :8*__av_get_bytes_per_sample(a->s.dec->sample_fmt),
#else
                  :__av_get_bits_per_sample_fmt(a->s.dec->sample_fmt),
#endif
          (double)a->s.dec->sample_rate,
          track->buf->bps,
          (double)track->buf->rate,
          v->s.codec?v->s.codec->name:"undefined");
    }
    else {
      _snprintf(buf->formatinformation,
          sizeof buf->formatinformation,
          "Length: %d seconds"
          "\nDecoder: " DECODER
          "\n\nAudio Codec: %s"
          "\nChannels: %d"
          "\nBits per sample (in): %d"
          "\nSample Rate (in): %.f Hz"
          "\nBits per sample (out): %d"
          "\nSample Rate (out): %.f Hz"
          ,
          buf->length/1000,
          a->s.codec?a->s.codec->name:"undefinded",
          a->s.dec->channels,
          0<a->s.dec->bits_per_raw_sample
                  ?a->s.dec->bits_per_raw_sample
#ifndef OLD
                  :8*__av_get_bytes_per_sample(a->s.dec->sample_fmt),
#else
                  :__av_get_bits_per_sample_fmt(a->s.dec->sample_fmt),
#endif
          (double)a->s.dec->sample_rate,
          track->buf->bps,
          (double)track->buf->rate);
    }
  }
}

static void get_tags_fn(samplebuf_t *buf)
{
  char tmp[256];
  char *mp=tmp+sizeof(tmp)-1;
  const char *fn=buf->fn+strlen(buf->fn);
  int offs;
  FILE *csv=NULL;
  int ch;

  while (buf->fn<fn&&'\\'!=fn[-1]&&'/'!=fn[-1])
    --fn;

  offs=fn - buf->fn;
  memcpy(buf->csv, buf->fn, offs);
  strcpy(buf->csv+offs, "folder.csv");

  if (NULL==(csv=fopen(buf->csv, "rb")))
    return;

  while ('\n'!=(ch=fgetc(csv))) {
    if (EOF==ch)
      goto cleanup;
  }

  for (;;) {
    char *wp=tmp;

    // File
    while ('\t'!=(ch=fgetc(csv))) {
      if (EOF==ch||wp==mp)
        goto cleanup;
      else
        *wp++=ch;
    }

    *wp=0;

    if (0==strcmp(fn, tmp)) {
#define MAKE_INFO(id) \
      wp=buf->tag.id; \
      mp=buf->tag.id+sizeof(buf->tag.id)-1; \
 \
      for (;;) { \
        ch=fgetc(csv); \
 \
        if (wp+1==mp) \
          goto cleanup; \
        else if ('\t'==ch||'\n'==ch||EOF==ch) \
          break; \
        else \
          *wp++=ch; \
      } \
 \
      *wp=0;
#define MAKE_NEW_INFO(id)
#define MAKE_ALTERNATE_INFO(id1, id2)
#include "info.h"
#undef MAKE_ALTERNATE_INFO
#undef MAKE_NEW_INFO
#undef MAKE_INFO
      break;
    }
    else {
      while ('\n'!=(ch=fgetc(csv))) {
        if (EOF==ch)
          goto cleanup;
      }
    }
  }
cleanup:
  if (NULL!=csv)
    fclose(csv);
}

#ifdef __SYNC__
static void in_ffsox_sync(track_t *track, int *guard)
{
  video_t *v=&track->v;

  if (NULL!=v->s.codec && 0==*guard) {
    EnterCriticalSection(&v->sync.cs);

    if (++v->sync.count<2) {
      LeaveCriticalSection(&v->sync.cs);
      WaitForSingleObject(&v->sync.hEvent, INFINITE);
      EnterCriticalSection(&v->sync.cs);
    }
    else
      SetEvent(&v->sync.hEvent);

    LeaveCriticalSection(&v->sync.cs);
    *guard=1;
  }
}
#endif // __SYNC__

static void in_ffsox_decode_video(track_t *track, AVPacket *p1)
{
  audio_t *a=&track->a;
  video_t *v=&track->v;
  AVFrame *frame=v->frame;
  int got_picture=0;
  int64_t pts;

  v->s.dec->reordered_opaque=p1->pts;
  __avcodec_decode_video2(v->s.dec, frame, &got_picture, p1);

  if (got_picture) {
    if (NULL==v->img_convert_ctx) {
      if (! v->frame->interlaced_frame) {
        __av_picture_copy(&v->out_pict, (AVPicture *)v->frame,
            v->in_pix_fmt, v->width, v->height);
      }
      else {
        __avpicture_deinterlace(&v->out_pict, (AVPicture *)v->frame,
            v->in_pix_fmt, v->width, v->height);
      }
    }
    else if (! v->frame->interlaced_frame) {
      __sws_scale(v->img_convert_ctx, v->frame->data, v->frame->linesize,
          0, v->height, v->out_pict.data, v->out_pict.linesize);
    }
    else {
      __avpicture_deinterlace(&v->in_pict, (AVPicture *)v->frame,
          v->in_pix_fmt, v->width, v->height);
      __sws_scale(v->img_convert_ctx, v->in_pict.data, v->in_pict.linesize,
          0, v->height, v->out_pict.data, v->out_pict.linesize);
    }

#ifndef OLD
    pts=*(int64_t*)__av_opt_ptr(__avcodec_get_frame_class(),
	    frame,"best_effort_timestamp");

    if (pts==AV_NOPTS_VALUE)
      pts=0;
#else // OLD
    if (p1->dts!=AV_NOPTS_VALUE) {
      pts=p1->dts;
#ifdef __DEBUG__
//#define __XDEBUG__
#endif
#ifdef __XDEBUG__
      fprintf(stderr, "!!! 1: pts: %.0f, %.0f\n", (double)pts,
          (double)track->start_time);
#endif
    }
    else if (v->frame->reordered_opaque!=AV_NOPTS_VALUE) {
      pts=v->frame->reordered_opaque;
#ifdef __XDEBUG__
      fprintf(stderr, "!!! 2: pts: %.0f\n", (double)pts);
#endif
    }
    else {
      pts=v->pts;
      v->pts+=v->step;
#ifdef __XDEBUG__
      fprintf(stderr, "!!! 3: pts: %.0f\n", (double)pts);
#endif
    }
#endif // OLD

#ifdef __SYNC__
    in_ffsox_sync(track, &v->sync.v);
#endif // __SYNC__
    double delay=v->s.time_base*(pts-track->start_time) - v->gettime();

	if (! ini.plugin.drop_frames.on
	    || (ini.plugin.drop_frames.before<=delay
	    && delay<=ini.plugin.drop_frames.after)) {

      if (!a->done&&0.0<delay)
        timer_sleep(&v->timer, delay);

      vout_draw(v->vout,
          v->out_pict.data[0], v->out_pict.linesize[0],
          v->out_pict.data[2], v->out_pict.linesize[2],
          v->out_pict.data[1], v->out_pict.linesize[1]);
	}
  }
}

#ifndef OLD
static void in_ffsox_decode_audio(track_t *track)
{
  audio_t *a=&track->a;
  AVPacket *p2=&track->a.p2;
  AVFrame *frame=a->frame;
  int got_frame=0;
  int n=__avcodec_decode_audio4(a->s.dec,frame,&got_frame,p2);

  if (n<0) {
    p2->size=0;
    p2->data=NULL;
  }
  else {
	if (got_frame) {
      int data_size=__av_samples_get_buffer_size(NULL,a->s.dec->channels,
          a->frame->nb_samples,a->s.dec->sample_fmt,1);

      a->rp=frame->data[0];
      a->mp=a->rp+data_size;
	}

    p2->size-=n;
    p2->data+=n;
/*
    fwrite(ff_samples, frame_size, 1, stdout);
*/
  }
}
#else // OLD
static void in_ffsox_decode_audio(track_t *track, decode_t decode)
{
  audio_t *a=&track->a;
  AVPacket *p2=&track->a.p2;
  int frame_size = sizeof(ff_samples);
  int n=decode(a->s.dec, (int16_t *)ff_samples, &frame_size, p2);

  if (n<0) {
    p2->size=0;
    p2->data=NULL;
  }
  else {
    p2->size-=n;
    p2->data+=n;
/*
    fwrite(ff_samples, frame_size, 1, stdout);
*/
    a->rp=ff_samples;
    a->mp=ff_samples+frame_size;
  }
}
#endif // OLD

static int in_ffsox_convert_audio(sox_effect_t * effp, sox_sample_t *samples,
    size_t len)
{
  track_t *track=*(track_t **)effp->priv;
  audio_t *a=&track->a;
  video_t *v=&track->v;
#ifndef OLD
  int bytes=__av_get_bytes_per_sample(a->s.dec->sample_fmt);
  int bps=8*bytes;
#else
  int bps=__av_get_bits_per_sample_fmt(a->s.dec->sample_fmt);
  int bytes=bps/8;
#endif
  int avail=(a->mp-a->rp)/bytes;
  sox_sample_t *wp=samples;
  sox_sample_t *mp=samples+(len<avail?len:avail);

  if (AV_SAMPLE_FMT_FLT==a->s.dec->sample_fmt) {
    double q=track->buf->tag.q;
	SOX_SAMPLE_LOCALS;

    while (wp<mp) {
	  double x=*(float *)a->rp;

	  if (0==a->apply_gain)
	    x*=q;

      *wp++=SOX_FLOAT_64BIT_TO_SAMPLE(x,effp->clips);
      a->rp+=bytes;
    }
  }
  else {
    while (wp<mp) {
      *wp++=16==bps?SOX_SIGNED_16BIT_TO_SAMPLE(*(int16_t *)a->rp,effp->clips):
          SOX_SIGNED_32BIT_TO_SAMPLE(*(int32_t *)a->rp,effp->clips);
      a->rp+=bytes;
    }
  }

#ifdef __SYNC__
  in_ffsox_sync(track, &v->sync.a);
#endif // __SYNC__

  return wp-samples;
}

static int in_ffsox_read_frame_a(track_t *track, AVPacket *p1)
{
   return __av_read_frame(track->ic, p1);
}

static int in_ffsox_read_frame_v(track_t *track, AVPacket *p1)
{
   return packet_queue_read(&track->v.qa, p1)?0:-1;
}

static int in_ffsox_is_seekable(track_t *track)
{
  if (ini.plugin.force_seek)
    return 1;
  else {
    int seekable=NULL!=track->ic->iformat->read_seek2;

#if FF_API_READ_SEEK
    seekable=seekable||NULL!=track->ic->iformat->read_seek;
#endif

    return seekable;
  }
}

#ifndef OLD
static int in_ffsox_read(sox_effect_t * effp, sox_sample_t *samples, size_t len)
{
  track_t *track=*(track_t **)effp->priv;
  audio_t *a=&track->a;
  AVPacket *p1=&a->p1;
  AVPacket *p2=&a->p2;

  for (;;) {
    if (a->rp<a->mp)
      return in_ffsox_convert_audio(effp, samples, len);
    else if (p2->size<0||a->eof)
      return 0;
    else if (0==p2->size) {
      if (NULL!=p1->data) {
        __av_free_packet(p1);
        memset(p1, 0, sizeof *p1);
      }

      if (a->read_frame(track, p1)<0) {
        a->eof=1;
#ifndef OLD
        return 0;
#else
        if (NULL==a->s.codec->decode)
          return 0;
		else if (__mpc7_decoder==a->s.codec)
		  return 0;
        else
          in_ffsox_decode_audio(track, a->s.codec->decode);
#endif
      }
      else if(p1->stream_index==a->s.id) {
        if (!a->start) {
		  if (AV_NOPTS_VALUE!=p1->pts) {
            int time=a->s.time_base*(p1->pts-track->start_time);

			if (0<time)
		      a->settime(time);
          }

		  a->start=1;
        }

        p2->data=p1->data;
        p2->size=p1->size;
      }
      else {
        __av_free_packet(p1);
        memset(p1, 0, sizeof *p1);
      }
    }
    else
#ifndef OLD
      in_ffsox_decode_audio(track);
#else
      in_ffsox_decode_audio(track, (decode_t)__avcodec_decode_audio3);
#endif
  }
}
#else // OLD
static int in_ffsox_read(sox_effect_t * effp, sox_sample_t *samples, size_t len)
{
  track_t *track=*(track_t **)effp->priv;
  audio_t *a=&track->a;
  AVPacket *p1=&a->p1;
  AVPacket *p2=&a->p2;

  for (;;) {
    if (a->rp<a->mp)
      return in_ffsox_convert_audio(effp, samples, len);
    else if (p2->size<0||a->eof)
      return 0;
    else if (0==p2->size) {
      if (NULL!=p1->data) {
        __av_free_packet(p1);
        memset(p1, 0, sizeof *p1);
      }

      if (a->read_frame(track, p1)<0) {
        a->eof=1;

        if (NULL==a->s.codec->decode)
          return 0;
		else if (__mpc7_decoder==a->s.codec)
		  return 0;
        else
          in_ffsox_decode_audio(track, a->s.codec->decode);
      }
      else if(p1->stream_index==a->s.id) {
        if (!a->start) {
		  if (AV_NOPTS_VALUE!=p1->pts) {
            int time=a->s.time_base*(p1->pts-track->start_time);

			if (0<time)
		      a->settime(time);
          }

		  a->start=1;
        }

        p2->data=p1->data;
        p2->size=p1->size;
      }
      else {
        __av_free_packet(p1);
        memset(p1, 0, sizeof *p1);
      }
    }
    else
      in_ffsox_decode_audio(track, (decode_t)__avcodec_decode_audio3);
  }
}
#endif // OLD

/* The function that will be called to input samples into the effects chain.
 * In this example, we get samples to process from a SoX-openned audio file.
 * In a different application, they might be generated or come from a different
 * part of the application. */
static int input_drain(
    sox_effect_t * effp, sox_sample_t * obuf, size_t * osamp)
{
  //(void)effp;   /* This parameter is not needed in this example */

  /* ensure that *osamp is a multiple of the number of channels. */
  *osamp -= *osamp % effp->out_signal.channels;

  /* Read up to *osamp samples into obuf; store the actual number read
   * back to *osamp */
  *osamp = in_ffsox_read(effp, obuf, *osamp);

  /* sox_read may return a number that is less than was requested; only if
   * 0 samples is returned does it indicate that end-of-file has been reached
   * or an error has occurred */
/*
  if (!*osamp && in->sox_errno)
    fprintf(stderr, "%s: %s\n", in->filename, in->sox_errstr);
*/

  return *osamp? SOX_SUCCESS : SOX_EOF;
}

/* A `stub' effect handler to handle inputting samples to the effects
 * chain; the only function needed for this example is `drain' */
static sox_effect_handler_t const * input_handler(void)
{
  static sox_effect_handler_t handler = {
    "input", NULL, SOX_EFF_MCHAN, NULL, NULL, NULL, input_drain, NULL, NULL, sizeof(track_t *)
  };
  return &handler;
}

static DWORD WINAPI in_ffsox_audio_thread(LPVOID p)
{
  track_t *track=p;

  __sox_flow_effects(track->a.chain, NULL, NULL);
  track->a.done=1;
  samplebuf_signal_done(track->buf);

  return 0;
}

static DWORD WINAPI in_ffsox_video_thread(LPVOID p)
{
  track_t *track=p;
  video_t *v=&track->v;
  AVPacket pkt;

  while (NULL!=packet_queue_read(&v->qv, &pkt)) {
    in_ffsox_decode_video(track, &pkt);
    __av_free_packet(&pkt);
  }

  return 0;
}

static DWORD WINAPI in_ffsox_dispatch_thread(LPVOID p)
{
  track_t *track=p;
  video_t *v=&track->v;
  audio_t *a=&track->a;
  packet_queue_t *qa=&v->qa;
  packet_queue_t *qv=&v->qv;
  AVPacket p1;

  HANDLE hThreadVideo=CreateThread(
    NULL,               // lpThreadAttributes (0 / CREATE_SUSPENDED)
    0,                  // dwStackSize
    in_ffsox_video_thread,    // lpStartAddress
    track,              // lpParameter
    0,                  // dwCreationFlags
    NULL                // lpThreadId
  );

  HANDLE hThreadAudio=CreateThread(
    NULL,               // lpThreadAttributes (0 / CREATE_SUSPENDED)
    0,                  // dwStackSize
    in_ffsox_audio_thread,    // lpStartAddress
    track,              // lpParameter
    0,                  // dwCreationFlags
    NULL                // lpThreadId
  );

#ifdef __DEBUG__
  fprintf(stderr, "start dispatching ...\n");
#endif
  for (;;) {
    for (;;) {
	  int of=packet_queue_overflow(qa)||packet_queue_overflow(qv);
	  int uf=packet_queue_underflow(qa)||packet_queue_underflow(qv);
	  int wait=of&&!uf;

	  wait=wait||v->pause;
	  wait=wait&&v->run;

	  if (wait)
        WaitForSingleObject(v->hEventWrite, INFINITE);
	  else
	    break;
	}

    if (! v->run)
      break;
    else if (__av_read_frame(track->ic, &p1)<0)
      v->run=0;
    else if(p1.stream_index==a->s.id)
      packet_queue_write(qa, &p1);
    else if(p1.stream_index==v->s.id)
      packet_queue_write(qv, &p1);
    else
      __av_free_packet(&p1);
  }

  packet_queue_signal_eof(qa);
  packet_queue_signal_eof(qv);
#ifdef __DEBUG__
  fprintf(stderr, "stop dispatching ...\n");
#endif
#ifdef __DEBUG__
  fprintf(stderr, "waiting for audio ...\n");
#endif
  WaitForSingleObject(hThreadAudio, INFINITE);
  CloseHandle(hThreadAudio);

#ifdef __DEBUG__
  fprintf(stderr, "waiting for video ...\n");
#endif
  WaitForSingleObject(hThreadVideo, INFINITE);
  CloseHandle(hThreadVideo);

#ifdef __DEBUG__
  fprintf(stderr, "done ...\n");
#endif
  return 0;
}

static void in_ffsox_close(track_t *track);

#ifdef __FLOAT__
static AVCodec *in_ffsox_find_decoder(enum CodecID id)
{
#ifdef __ZZZ__
    AVCodec *p=__av_codec_next(NULL);
	AVCodec *p1=NULL;

    while (NULL!=p) {
        if (p->decode!=NULL&&p->id==id)
            return p;

        p=__av_codec_next(p);
    }

    return NULL;
#else
  if (CODEC_ID_AC3==id) {
#ifdef __CHOOSE_LIB__
    extern int a52_stub_valid;
    AVCodec *p;

#ifdef __DEBUG__
    fprintf(stderr, "AC3: \"%s\" ...\n",ini.plugin._ac3_decoder);
#endif

    if (!strcmp(LIBA52,ini.plugin._ac3_decoder)) {
	  if (NULL!=(p=__avcodec_find_decoder_by_name("liba52")))
	    return p;
	}
#endif

    return __avcodec_find_decoder(id);
  }
  else if (CODEC_ID_MP1==id) {
    AVCodec *p;

	if (! ini.plugin.prefer_float)
      return __avcodec_find_decoder(id);
    else if (NULL!=(p=__avcodec_find_decoder_by_name("mp1float")))
	  return p;
    else
      return __avcodec_find_decoder(id);
  }
  else if (CODEC_ID_MP2==id) {
    AVCodec *p;

	if (! ini.plugin.prefer_float)
      return __avcodec_find_decoder(id);
    else if (NULL!=(p=__avcodec_find_decoder_by_name("mp2float")))
	  return p;
    else
      return __avcodec_find_decoder(id);
  }
  else if (CODEC_ID_MP3==id) {
#ifndef __CHOOSE_LIB__
    AVCodec *p;
#else
    extern int mad_stub_valid;
    AVCodec *p;

    if (!strcmp(LIBMAD,ini.plugin._mp3_decoder)) {
	  if (NULL!=(p=__avcodec_find_decoder_by_name("libmad")))
	    return p;
	}
#endif

	if (! ini.plugin.prefer_float)
      return __avcodec_find_decoder(id);
    else if (NULL!=(p=__avcodec_find_decoder_by_name("mp3float")))
	  return p;
    else
      return __avcodec_find_decoder(id);
  }
  else
    return __avcodec_find_decoder(id);
#endif
}
#endif // __FLOAT__

static void in_ffsox_set_dyn_range(track_t *track)
{
  audio_t *a=&track->a;
  AVCodecContext *dc=a->s.dec;
  const AVClass *priv_class=a->s.codec->priv_class;
  char *priv_data=dc->priv_data;
  const AVOption *option;

  if (priv_class&&priv_data) {
    for (option=priv_class->option;option&&option->name;++option) {
      if (0==strcmp("drc_scale",option->name)) {
	     if (FF_OPT_TYPE_FLOAT==option->type) {
           *(float *)(((char *)dc->priv_data)+option->offset)=0.0;
	     }

        break;
      }
	}
  }
}

static track_t *in_ffsox_open(track_t *track, samplebuf_t *buf, const char *fn)
{
  audio_t *a=&track->a;
  video_t *v=&track->v;
  AVFormatContext *ic=NULL;
  int id;

  memset(track, 0, sizeof *track);
  track->a.s.id=-1;
  track->v.s.id=-1;

#ifdef __DEBUG__
  fprintf(stderr, "*** Going to open AVFormat: \"%s\".\n",fn);
#endif

#ifndef OLD
  if (0!=__avformat_open_input(&ic, fn, NULL, NULL))
    return NULL;
#else
  if (0!=__av_open_input_file(&ic, fn, NULL, 0, NULL))
    return NULL;
#endif

  track->ic=ic;

#ifndef OLD
  if (__avformat_find_stream_info(ic,NULL)<0)
    goto error;
#else
  if (__av_find_stream_info(ic)<0)
    goto error;
#endif

  for (id=0;id<ic->nb_streams;++id) {
    AVStream *stream=ic->streams[id];

    if (stream->codec->codec_type==AVMEDIA_TYPE_AUDIO) {
	  if (a->s.id<0||2==stream->codec->channels) {
        a->s.id=id;
        a->s.stream=stream;
        a->s.dec=stream->codec;
#ifdef __FLOAT__
		//a->s.dec->request_channel_layout=AV_CH_LAYOUT_STEREO_DOWNMIX;
		//a->s.dec->request_channel_layout=AV_CH_LAYOUT_STEREO;
		a->s.dec->request_channels=2;

		if (ini.plugin.prefer_float)
		  a->s.dec->request_sample_fmt=AV_SAMPLE_FMT_FLT;
#endif // __FLOAT__
        a->s.codec=in_ffsox_find_decoder(a->s.dec->codec_id);
#if defined (__FLOAT__) && defined (__DEBUG__)
fprintf(stderr, "!!! AV_SAMPLE_FMT_FLT: %d !!!!!!!!!!!!!\n",AV_SAMPLE_FMT_FLT==a->s.dec->sample_fmt);
#endif // __FLOAT__
	  }
    }
    else if (stream->codec->codec_type==AVMEDIA_TYPE_VIDEO
	    &&ini.winamp.vid) {

	  if (v->s.id<0) {
        v->s.id=id;
        v->s.stream=stream;
        v->s.dec=stream->codec;
        v->s.codec=in_ffsox_find_decoder(v->s.dec->codec_id);
	  }
    }
  }

  if (NULL==a->s.codec)
    goto error;

  samplebuf_open(buf, &ffsox_plugin, track, fn, a->s.dec->channels,
      a->s.dec->sample_rate,
#ifndef OLD
	  8*__av_get_bytes_per_sample(a->s.dec->sample_fmt),
#else
	  __av_get_bits_per_sample_fmt(a->s.dec->sample_fmt),
#endif
      ic->duration/1000);
  track->buf=buf;

  if (NULL!=ic->metadata)
    get_tags_md(ic->metadata, buf);

/*
  if (NULL!=a->s.stream->metadata)
    get_tags_md(a->s.stream->metadata, buf);
*/

#ifndef __OPEN_LATE__
  a->start=0;
  a->done=0;
  //a->settime=settime;
  //samplebuf_start(track->buf);
#ifndef OLD
  __avcodec_open2(a->s.dec,a->s.codec,NULL);
  a->frame=__avcodec_alloc_frame();
  __avcodec_get_frame_defaults(a->frame);
#else
  __avcodec_open(a->s.dec, a->s.codec);
#endif
  in_ffsox_set_dyn_range(track);
  a->s.time_base=1000.0*av_q2d(a->s.stream->time_base);
  a->apply_gain=1;
#endif // __OPEN_LATE__

  get_tags_fn(buf);
  prefs_set_gain(&buf->tag);
  in_ffsox_fill_format(track);

  return track;
error:
  in_ffsox_close(track);

  return NULL;
}

void liba52_adapter_set_gain(AVCodecContext *avctx, double gain);
void libmad_adapter_set_gain(AVCodecContext *avctx, double gain);

static track_t *in_ffsox_start(track_t *track, vout_t *vout, int (*gettime)(),
    void (*settime)(int), void (*showvis)())
{
  audio_t *a=&track->a;
  video_t *v=&track->v;
  DWORD WINAPI (*thread)(LPVOID p);

#ifdef __OPEN_LATE__
  a->start=0;
  a->done=0;
  a->settime=settime;
  samplebuf_start(track->buf);
#ifndef OLD
  __avcodec_open2(a->s.dec,a->s.codec,NULL);
  a->frame=__avcodec_alloc_frame();
  __avcodec_get_frame_defaults(a->frame);
#else
  __avcodec_open(a->s.dec, a->s.codec);
#endif
  in_ffsox_set_dyn_range(track);
  a->s.time_base=1000.0*av_q2d(a->s.stream->time_base);
  a->apply_gain=1;
#else
  a->settime=settime;
  samplebuf_start(track->buf);
#endif // __OPEN_LATE__

  if (!strcmp("liba52",a->s.codec->name)) {
    //double gain=samplebuf_get_gain(track->buf);
    double gain=track->buf->tag.gain;

	if (gain<0.0) {
      double q=pow(2.0,gain/6.0);
      liba52_adapter_set_gain(a->s.dec,q);
      a->apply_gain=0;
    }
  }
  else if (!strcmp("libmad",a->s.codec->name)) {
    //double gain=samplebuf_get_gain(track->buf);
    double gain=track->buf->tag.gain;

	if (gain<0.0) {
      double q=pow(2.0,gain/6.0);
      libmad_adapter_set_gain(a->s.dec,q);
      a->apply_gain=0;
    }
  }
#ifdef __FLOAT__
  else if (AV_SAMPLE_FMT_FLT==a->s.dec->sample_fmt) {
#ifdef __DEBUG__
fprintf(stderr, "!!! FLOAT !!!!!!!!!!!!!!!!!!!\n");
#endif // __DEBUG__
    if (track->buf->tag.gain<0.0)
      a->apply_gain=0;
  }
#endif // __FLOAT__

  if (NULL!=v->s.codec&&ini.winamp.vid) {
    AVStream *stream=v->s.stream;
    AVCodecContext *dec=v->s.dec;
    double aspectratio1, aspectratio2;

#ifndef OLD
    __avcodec_open2(v->s.dec,v->s.codec,NULL);
#else
    __avcodec_open(v->s.dec, v->s.codec);
#endif

    v->hEventWrite=CreateEvent(
      NULL,           // lpEventAttributes
      FALSE,          // bManualReset
      FALSE,          // bInitialState
      NULL            // lpName
    );

    v->run=1;
    v->pause=0;
    packet_queue_open(&v->qa, ini.queues.audio.min, ini.queues.audio.max,
	    v->hEventWrite);
    packet_queue_open(&v->qv, ini.queues.video.min, ini.queues.video.max,
	    v->hEventWrite);
    a->read_frame=in_ffsox_read_frame_v;
    thread=in_ffsox_dispatch_thread;

    ///////////////////////////////////////////////////////////////////////////
    v->width=v->s.dec->width;
    v->height=v->s.dec->height;
    v->in_pix_fmt=v->s.dec->pix_fmt;
    v->out_pix_fmt=PIX_FMT_YUV420P;

    aspectratio1
        =stream->sample_aspect_ratio.num?av_q2d(stream->sample_aspect_ratio)
        :dec->sample_aspect_ratio.num?av_q2d(dec->sample_aspect_ratio)
        :0.0
        ;
    aspectratio2=aspectratio1<=0.0?1.0:aspectratio1;

    vout_open(vout, v->width, v->height, 1.0/aspectratio2);
    vout_set_caption(vout, FFSOX);
    v->vout=vout;
//vfs();

    v->s.time_base=1000.0*av_q2d(stream->time_base);

	track->start_time=a->s.stream->start_time<v->s.stream->start_time
	    ?a->s.stream->start_time
	    :v->s.stream->start_time
		;
    v->step=1000.0/av_q2d(stream->r_frame_rate);
    v->gettime=gettime;
    timer_open(&v->timer, ini.plugin.timer_res);

    v->frame= __avcodec_alloc_frame();

    if (v->out_pix_fmt!=v->in_pix_fmt) {
      v->img_convert_ctx=__sws_getContext(
          v->width, v->height, v->in_pix_fmt,
          v->width, v->height, v->out_pix_fmt,
          SWS_BICUBIC, NULL, NULL, NULL);
    }

    __avpicture_alloc(&v->out_pict, v->out_pix_fmt, v->width, v->height);
    __avpicture_alloc(&v->in_pict, v->in_pix_fmt, v->width, v->height);

#ifdef __SYNC__
    // sync ////////////////////////////////
    InitializeCriticalSection(&v->sync.cs);

    v->sync.hEvent=CreateEvent(
      NULL,           // lpEventAttributes
      TRUE,           // bManualReset
      FALSE,          // bInitialState
      NULL            // lpName
    );

    v->sync.count=0;
    v->sync.a=0;
    v->sync.v=0;
#endif // __SYNC__
  }
  else {
	track->start_time=a->s.stream->start_time;
    a->read_frame=in_ffsox_read_frame_a;
    thread=in_ffsox_audio_thread;

    if (ini.winamp.vis)
      showvis();
  }

  a->chain=track_init_sox_chain(track, track->buf,
#ifndef OLD
      8*__av_get_bytes_per_sample(a->s.dec->sample_fmt),
#else
      __av_get_bits_per_sample_fmt(a->s.dec->sample_fmt),
#endif
	  a->s.dec->sample_rate, a->s.dec->channels,
	  a->apply_gain, input_handler);

  track->hThreadSoX=CreateThread(
    NULL,       // lpThreadAttributes (0 / CREATE_SUSPENDED)
    0,          // dwStackSize
    thread,     // lpStartAddress
    track,      // lpParameter
    0,          // dwCreationFlags
    NULL        // lpThreadId
  );

  return track;
}

static void in_ffsox_close(track_t *track)
{
#ifndef __OPEN_LATE__
  audio_t *a=&track->a;
#endif // __OPEN_LATE__

  samplebuf_close(track->buf);

  if (NULL!=track->ic) {
#ifndef OLD
    __avformat_close_input(&track->ic);
#else
    __av_close_input_file(track->ic);
#endif
    track->ic=NULL;
#ifdef __DEBUG__
	fprintf(stderr, "*** AVFormat closed.\n");
#endif
  }

  memset(track, 0, sizeof *track);

#ifndef __OPEN_LATE__
  if (NULL!=a->s.codec) {
#ifndef OLD
   __av_free(a->frame);
#endif
    __avcodec_close(a->s.dec);
    a->s.dec=NULL;
    a->s.codec=NULL;
  }
#endif // __OPEN_LATE__

}

static void in_ffsox_stop(track_t *track)
{
  audio_t *a=&track->a;
  video_t *v=&track->v;

  if (NULL!=track->hThreadSoX) {
    WaitForSingleObject(track->hThreadSoX, INFINITE);
    CloseHandle(track->hThreadSoX);
  }

  if (NULL!=track->a.p1.data) {
    __av_free_packet(&track->a.p1);
    memset(&track->a.p1, 0, sizeof track->a.p1);
  }

  if (NULL!=a->chain)
    __sox_delete_effects_chain(a->chain);

  if (NULL!=v->s.codec) {
#ifdef __SYNC__
    if (NULL!=v->sync.hEvent)
      CloseHandle(v->sync.hEvent);

    DeleteCriticalSection(&v->sync.cs);
#endif // __SYNC__

    __avpicture_free(&v->in_pict);
    __avpicture_free(&v->out_pict);

    if (NULL!=v->img_convert_ctx)
      __sws_freeContext(v->img_convert_ctx);

    if (NULL!=track->v.frame)
      __av_free(track->v.frame);

    timer_close(&track->v.timer);

    if (NULL!=track->v.vout)
      vout_close(track->v.vout);

    packet_queue_close(&v->qv);
    packet_queue_close(&v->qa);

    if (NULL!=v->hEventWrite)
      CloseHandle(v->hEventWrite);

    if (NULL!=v->s.codec) {
      __avcodec_close(v->s.dec);
      v->s.dec=NULL;
      v->s.codec=NULL;
    }
  }

#ifdef __OPEN_LATE__
  if (NULL!=a->s.codec) {
#ifndef OLD
   __av_free(a->frame);
#endif
    __avcodec_close(a->s.dec);
    a->s.dec=NULL;
    a->s.codec=NULL;
  }
#endif // __OPEN_LATE__

  samplebuf_stop(track->buf);
}

static int filetype(const char *path, const char *ext)
{
  const char *rp=path+strlen(path);

  while (path<rp&&'.'!=rp[-1])
    --rp;

  return 0==strcmpi(rp,ext);
}

static track_t *in_ffsox_create_play(samplebuf_t *buf, const char *fn) {
#ifndef __FLAC_SOX__
  return ffmpeg_stub_valid?in_ffsox_open(&track_play, buf, fn):NULL;
#else // __FLAC_SOX__
  return filetype(fn, "flac")
    ?sox_plugin.create_play(buf, fn)
    :in_ffsox_open(&track_play, buf, fn)
    ;
#endif // __FLAC_SOX__
}

static track_t *in_ffsox_create_info(samplebuf_t *buf, const char *fn) {
#ifndef __FLAC_SOX__
  return ffmpeg_stub_valid?in_ffsox_open(&track_info, buf, fn):NULL;
#else // __FLAC_SOX__
  return filetype(fn, "flac")
    ?sox_plugin.create_info(buf, fn)
    :in_ffsox_open(&track_info, buf, fn)
    ;
#endif // __FLAC_SOX__
}

static void in_ffsox_kill(track_t *track)
{
  audio_t *a=&track->a;
  video_t *v=&track->v;

  if (NULL!=v->s.codec) {
    v->run=0;
    packet_queue_signal_kill(&v->qa);
    packet_queue_signal_kill(&v->qv);
    timer_kill(&v->timer);
  }

  samplebuf_signal_done(track->buf);
}

static int in_ffsox_seek(track_t *track, int pos_ms)
{
  if (in_ffsox_is_seekable(track)) {
    audio_t *a=&track->a;
    video_t *v=&track->v;
	stream_t *s=NULL!=v->s.codec?&v->s:&a->s;
    int64_t seek_target=__av_rescale_q(1000*pos_ms, AV_TIME_BASE_Q,
        s->stream->time_base);

    in_ffsox_kill(track);

    if (NULL!=track->hThreadSoX) {
      WaitForSingleObject(track->hThreadSoX, INFINITE);
      CloseHandle(track->hThreadSoX);
      track->hThreadSoX=NULL;
    }

#ifndef __OLD__
    return __avformat_seek_file(track->ic,s->id,INT64_MIN,seek_target,
		INT64_MAX,0)<0?0:1;
#else
    return __av_seek_frame(track->ic, s->id, seek_target, 0)<0?0:1;
#endif
  }
  else
    return 0;
}

static void in_ffsox_flush(track_t *track)
{
  audio_t *a=&track->a;
  video_t *v=&track->v;

  samplebuf_reset(track->buf);
  a->done=0;
  a->start=0;

  if (NULL!=v->s.codec) {
    v->run=1;
    v->pause=0;
    packet_queue_reset(&v->qa);
    packet_queue_reset(&v->qv);
  }
}

static void in_ffsox_restart(track_t *track)
{
  audio_t *a=&track->a;
  video_t *v=&track->v;
  DWORD WINAPI (*thread)(LPVOID)=NULL!=v->s.codec
      ?in_ffsox_dispatch_thread
      :in_ffsox_audio_thread
	  ;

  track->hThreadSoX=CreateThread(
    NULL,       // lpThreadAttributes (0 / CREATE_SUSPENDED)
    0,          // dwStackSize
    thread,     // lpStartAddress
    track,      // lpParameter
    0,          // dwCreationFlags
    NULL        // lpThreadId
  );
}

static void in_ffsox_pause(track_t *track)
{
  video_t *v=&track->v;

  if (NULL!=v->s.codec) {
    v->qa.state=PACKET_QUEUE_PAUSE;
    v->qv.state=PACKET_QUEUE_PAUSE;
  }
}

static void in_ffsox_unpause(track_t *track)
{
  video_t *v=&track->v;

  if (NULL!=v->s.codec) {
    packet_queue_signal_work(&v->qa);
    packet_queue_signal_work(&v->qv);
  }
}

plugin_t ffsox_plugin = {
  .create_play=in_ffsox_create_play,
  .create_info=in_ffsox_create_info,
  .close=in_ffsox_close,
  .start=in_ffsox_start,
  .stop=in_ffsox_stop,
  .kill=in_ffsox_kill,
  .pause=in_ffsox_pause,
  .unpause=in_ffsox_unpause,
  .seek=in_ffsox_seek,
  .flush=in_ffsox_flush,
  .restart=in_ffsox_restart
};
