/*
 * main.c
 * Copyright (C) 2010 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <windows.h>
#include <Winamp/in2.h>
#define _MSC_VER 2000
#include <Winamp/wa_ipc.h>
#undef _MSC_VER
#include "plugin.h"
#include "strbuf.h"
#include "resources.h"
#define __DYNAMIC__
#include "libsox_stub.h"
#include "liba52_stub.h"
#include "libmad_stub.h"

// avoid CRT. Evil. Big. Bloated. Only uncomment this code if you are using 
// 'ignore default libraries' in VC++. Keeps DLL size way down.
// /*
BOOL WINAPI _DllMainCRTStartup(HANDLE hInst, ULONG ul_reason_for_call, LPVOID lpReserved)
{
	return TRUE;
}
// */

// post this to the main window at end of file (after playback as stopped)
#define WM_WA_MPEG_EOF WM_USER+2


// raw configuration.
CRITICAL_SECTION cs;

struct {
  CRITICAL_SECTION *buf;
  CRITICAL_SECTION *wp;
  CRITICAL_SECTION *mp;
} ffmpeg_cs;

strbuf_t fileinfo;
wstrbuf_t tag;

samplebuf_t pb;
samplebuf_t ib;
int getoutputtime();
void showvis();


In_Module mod;			// the output module (filled in near the bottom of this file)

char lastfn[3*MAX_PATH];	// currently playing file (used for getting info on the current file)

int decode_pos_ms;		// current decoding position, in milliseconds. 
						// Used for correcting DSP plug-in pitch changes
int paused;				// are we paused?
volatile int seek_needed; // if != -1, it is the point that the decode 
						  // thread should seek to, in ms.

//plugin_t *plugin=&sox_plugin;
//plugin_t *plugin=&ffsox_plugin;
int valid=0;

volatile int killDecodeThread=0;			// the kill switch for the decode thread
HANDLE thread_handle=INVALID_HANDLE_VALUE;	// the handle to the decode thread

static plugin_t *get_plugin(void)
  { return ini.plugin.sox_decoder?&sox_plugin:&ffsox_plugin; }

DWORD WINAPI DecodeThread(LPVOID b); // the decode thread procedure
int getoutputtime();
void settime(int ms);

void alert(char *msg)
{
  MessageBox(
    NULL,							// __in  HWND hWnd,
    msg,							// __in  LPCTSTR lpText,
    "Error Loading FFSoX-Player",	// __in  LPCTSTR lpCaption,
    MB_OK							// __in  UINT uType
  );
}

void config(HWND hwndParent)
{
	if (! valid)
		return;

	DialogBoxParam (GetModuleHandle(MODULE),
	    MAKEINTRESOURCE(IDD_DIALOG_PREFS),hwndParent,ConfigDlgProc,0);
//fprintf(stderr, "*** %d\n", GetLastError());
}

void about(HWND hwndParent)
{
	if (! valid)
		return;

	DialogBoxParam (GetModuleHandle(MODULE),
	    MAKEINTRESOURCE(IDD_DIALOG_ABOUT),hwndParent,AboutDlgProc,0);
}

int open_ffmpeg_locks()
{
  size_t size=8;

  if (NULL==(ffmpeg_cs.buf=malloc(size*(sizeof *ffmpeg_cs.buf))))
    goto error;

  ffmpeg_cs.wp=ffmpeg_cs.buf;
  ffmpeg_cs.mp=ffmpeg_cs.buf+size;

  return 0;
error:
  return -1;
}

void close_ffmpeg_locks()
{
  free(ffmpeg_cs.buf);
}

int ffmpeg_lock_callback(void **mutex, enum AVLockOp op)
{
  switch (op) {
    case AV_LOCK_CREATE:
      if (ffmpeg_cs.wp==ffmpeg_cs.mp) {
		size_t size=ffmpeg_cs.mp==ffmpeg_cs.buf;

        if (NULL==(ffmpeg_cs.buf=realloc(ffmpeg_cs.buf,
		    2*size*(sizeof *ffmpeg_cs.buf))))
	      goto error;

		ffmpeg_cs.wp=ffmpeg_cs.buf+size;
		ffmpeg_cs.mp=ffmpeg_cs.buf+2*size;
	  }

      InitializeCriticalSection(ffmpeg_cs.wp);
	  *mutex=ffmpeg_cs.wp;
	  ++ffmpeg_cs.wp;
#ifdef __DEBUG__
      fprintf(stderr,"AV_LOCK_CREATE: %p\n",*mutex);
#endif
	  break;
    case AV_LOCK_OBTAIN:
      EnterCriticalSection(*mutex);
#ifdef __DEBUG__
      fprintf(stderr,"AV_LOCK_OBTAIN: %p\n",*mutex);
#endif
	  break;
    case AV_LOCK_RELEASE:
#ifdef __DEBUG__
      fprintf(stderr,"AV_LOCK_RELEASE: %p\n",*mutex);
#endif
      LeaveCriticalSection(*mutex);
	  break;
    case AV_LOCK_DESTROY:
#ifdef __DEBUG__
      fprintf(stderr,"AV_LOCK_DESTROY: %p\n",*mutex);
#endif
      DeleteCriticalSection(*mutex);
	  break;
	default:
	  goto error;
  }

  return 0;
error:
  return -1;
}

void init() { 
#ifdef __EXTENSIONS__
	int i,j;
	char *ep=exts;
	sox_format_tab_t const *sox_format_fns;
#endif // __EXTENSIONS__
	/* any one-time initialization goes here (configuration reading, etc) */ 
#ifdef __DEBUG__
	AllocConsole();
	freopen("conin$", "r", stdin);
	freopen("conout$", "w", stdout);
	freopen("conout$", "w", stderr);
#endif
	load_ini();

	if (! setdlgitemurl_stub_load(ini.path, ini.p)) {
	  alert("Error loading \"in_ffsox.dll\": missing \"setdlgitemurl.dll\".");
	  goto error;
	}

	if (! sox_stub_load(ini.path, ini.p)) {
	  alert("Error loading \"in_ffsox.dll\": missing \"libsox.dll\".");
	  goto error;
    }

	__sox_init();

	if (ffmpeg_stub_load(ini.path, ini.p, alert)) {
	    open_ffmpeg_locks();
#ifdef __ALT__
		__av_register_all();
#else
		__avcodec_register_all();
		__av_register_all();
		__avformat_network_init();
		__av_lockmgr_register(ffmpeg_lock_callback);
#endif
		ffmpeg_find_decoders();

#ifdef __CHOOSE_LIB__
#ifdef __ALT__
		if (mad_stub_load(ini.path, ini.p)
		        &&!strcmp(LIBMAD,ini.plugin._mp3_decoder)
				&&NULL!=__mp3_decoder) {

			extern AVCodec libmad_adapter;

			libmad_adapter.next=__mp3_decoder->next;
			*__mp3_decoder=libmad_adapter;
		}

		if (a52_stub_load(ini.path, ini.p)
		        &&!strcmp(LIBA52,ini.plugin._ac3_decoder)
				&&NULL!=__ac3_decoder) {

			extern AVCodec liba52_adapter;

			liba52_adapter_start();
			liba52_adapter.next=__ac3_decoder->next;
			*__ac3_decoder=liba52_adapter;
		}
#else
		if (mad_stub_load(ini.path, ini.p)) {
			extern AVCodec libmad_adapter;

			__avcodec_register(&libmad_adapter);
		}

		if (a52_stub_load(ini.path, ini.p)) {
			extern AVCodec liba52_adapter;

			liba52_adapter_start();
			__avcodec_register(&liba52_adapter);
		}
#endif
#endif // __CHOOSE_LIB__
	}

	valid=1;

#ifdef __EXTENSIONS__
	sox_format_fns=__sox_get_format_fns();

	for (i=0;NULL!=sox_format_fns[i].fn;++i) {
		sox_format_handler_t const * handler = sox_format_fns[i].fn();

		for (j=0;NULL!=handler->names[j];++j) {
			const char *np=handler->names[j];
#ifdef __DEBUG__
			//fprintf(stderr, "--- %s\n", np);
#endif
			while (*np)
				*ep++=*np++;

      		*ep++=0;
      		*ep++='X';
			*ep++=0;
		}
	}

	if (ffmpeg_stub_valid) {
		AVInputFormat  *cur;

		for (cur=av_iformat_next(NULL);NULL!=cur;cur=av_iformat_next(cur)) {
			const char *np=cur->extensions;

			if (np) {
#ifdef __DEBUG__
				fprintf(stderr, "--- %s\n", np);
#endif
				while (*np)
					*ep++=*np++;

      			*ep++=0;
      			*ep++='X';
				*ep++=0;
			}
		}
	}

	*ep++=0;
#endif // __EXTENSIONS__
	mod.UsesOutputPlug |= IN_MODULE_FLAG_REPLAYGAIN;
	mod.UsesOutputPlug |= IN_MODULE_FLAG_REPLAYGAIN_PREAMP;  

    InitializeCriticalSection(&cs);
    strbuf_open(&fileinfo,3*MAX_PATH);
	wstrbuf_open(&tag,10245);

	return;
error:
	memset(exts, 0, sizeof exts);
}

void quit() { 
#ifdef __DEBUG__
    Sleep(2000);
#endif
	wstrbuf_close(&tag);
    strbuf_close(&fileinfo);
    DeleteCriticalSection(&cs);

    if (ffmpeg_stub_valid)
	  close_ffmpeg_locks();

	/* one-time deinit, such as memory freeing */ 
	if (sox_stub_valid)
	  __sox_quit();
}

int isourfile(const in_char *fn) { 
// used for detecting URL streams.. unused here. 
// return !strncmp(fn,"http://",7); to detect HTTP streams, etc
	return 0; 
} 

// called when winamp wants to play a file
int play(const in_char *wfilename) 
{ 
	int ret=1;
	int maxlatency;
	DWORD thread_id;
    strbuf_t filename;
	char *fn;

	EnterCriticalSection(&cs);
    strbuf_wchar2utf8(&fileinfo,ini.plugin.path_page,wfilename);
	fn=(char *)fileinfo.buf;

	paused=0;
	decode_pos_ms=0;
	seek_needed=-1;

	mod.UsesOutputPlug |= IN_MODULE_FLAG_REPLAYGAIN;
	mod.UsesOutputPlug |= IN_MODULE_FLAG_REPLAYGAIN_PREAMP;  
	
	// CHANGEME! Write your own file opening code here
	memset(&pb, 0, sizeof pb);
	get_plugin()->create_play(&pb, fn);
#ifdef __DEBUG__
	fprintf(stderr, "*** playing: \"%s\" ...\n", fn);
#endif

	if (NULL==pb.track) // error opening file
	{
		// we return error. 1 means to keep going in the playlist, -1
		// means to stop the playlist.
		goto cleanup;
	}
	
	strcpy(lastfn,fn);

	// -1 and -1 are to specify buffer and prebuffer lengths.
	// -1 means to use the default, which all input plug-ins should
	// really do.
#ifdef __DEBUG__
	fprintf(stderr, "*** playing: %d, %d, %d ...\n",pb.rate,pb.nch,pb.bps);
#endif
	maxlatency = mod.outMod->Open(pb.rate,pb.nch,pb.bps, -1,-1); 

	// maxlatency is the maxium latency between a outMod->Write() call and
	// when you hear those samples. In ms. Used primarily by the visualization
	// system.

	if (maxlatency < 0) // error opening device
	{
		pb.plugin->close(pb.track);
		pb.track=NULL;
		goto cleanup;
	}
	// dividing by 1000 for the first parameter of setinfo makes it
	// display 'H'... for hundred.. i.e. 14H Kbps.
	mod.SetInfo((pb.rate*pb.bps*pb.nch)/1000,pb.rate/1000,pb.nch,1);

	// initialize visualization stuff
	mod.SAVSAInit(maxlatency,pb.rate);
	mod.VSASetInfo(pb.rate,pb.nch);

	// set the output plug-ins default volume.
	// volume is 0-255, -666 is a token for
	// current volume.
	mod.outMod->SetVolume(-666); 

	// launch decode thread
	killDecodeThread=0;
	thread_handle = (HANDLE) 
		CreateThread(NULL,0,(LPTHREAD_START_ROUTINE) DecodeThread,NULL,0,&thread_id);
	
	ret=0;
cleanup:
	LeaveCriticalSection(&cs);

	return ret;
}

// standard pause implementation
void pause()
{
	//EnterCriticalSection(&cs);
	paused=1;
	mod.outMod->Pause(1);
	pb.plugin->pause(pb.track);
	//LeaveCriticalSection(&cs);
}

void unpause()
{
	//EnterCriticalSection(&cs);
	pb.plugin->unpause(pb.track);
	mod.outMod->Pause(0);
	paused=0;
	//LeaveCriticalSection(&cs);
}

int ispaused()
{
	return paused;
}


// stop playing.
void stop() { 
	//EnterCriticalSection(&cs);

	if (thread_handle != INVALID_HANDLE_VALUE)
	{
		if (0==killDecodeThread) {
		  killDecodeThread=1;
          pb.plugin->kill(pb.track); // simulte EOF in sox read
		}

		if (WaitForSingleObject(thread_handle,10000) == WAIT_TIMEOUT)
		{
			MessageBox(mod.hMainWindow,"error asking thread to die!\n",
				"error killing decode thread",0);
			TerminateThread(thread_handle,0);
		}
		CloseHandle(thread_handle);
		thread_handle = INVALID_HANDLE_VALUE;
	}

	// close output system
	mod.outMod->Close();

	// deinitialize visualization
	mod.SAVSADeInit();
	

	// CHANGEME! Write your own file closing code here
	if (NULL!=pb.track)
	{
		pb.plugin->close(pb.track);
		pb.track=NULL;
	}

	//LeaveCriticalSection(&cs);
}


// returns length of playing track
int getlength() {
	return pb.length;
}


// returns current output position, in ms.
// you could just use return mod.outMod->GetOutputTime(),
// but the dsp plug-ins that do tempo changing tend to make
// that wrong.
int getoutputtime() { 
	int outputtime;

	//EnterCriticalSection(&cs);
	outputtime=decode_pos_ms+
		(mod.outMod->GetOutputTime()-mod.outMod->GetWrittenTime()); 
	//LeaveCriticalSection(&cs);

	return outputtime;
}
void settime(int ms) {
	decode_pos_ms=ms;
}

// called when the user releases the seek scroll bar.
// usually we use it to set seek_needed to the seek
// point (seek_needed is -1 when no seek is needed)
// and the decode thread checks seek_needed.
void setoutputtime(int time_in_ms) { 
	seek_needed=time_in_ms; 
}


// standard volume/pan functions
void setvolume(int volume) { mod.outMod->SetVolume(volume); }
void setpan(int pan) { mod.outMod->SetPan(pan); }

// this gets called when the use hits Alt+3 to get the file info.
// if you need more info, ask me :)

int infoDlg(const in_char *fn, HWND hwnd)
{
	// CHANGEME! Write your own info dialog code here
	return 0;
}


// this is an odd function. it is used to get the title and/or
// length of a track.
// if filename is either NULL or of length 0, it means you should
// return the info of lastfn. Otherwise, return the information
// for the file in filename.
// if title is NULL, no title is copied into it.
// if length_in_ms is NULL, no length is copied into it.
void getfileinfo(const in_char *wfilename, in_char *title, int *length_in_ms)
{
    EnterCriticalSection(&cs);
    strbuf_wchar2utf8(&fileinfo,ini.plugin.path_page,wfilename);

	if (!*fileinfo.buf)  // currently playing file
	{
		if (length_in_ms) *length_in_ms=getlength();
		if (title) // get non-path portion.of filename
		{
#ifdef __ORIGINAL__
			char *p=lastfn+strlen(lastfn);
			while (*p != '\\' && p >= lastfn) p--;
			strcpy(title,++p);
#else
			char *p1=lastfn+strlen(lastfn);
			in_char *p2=title;

			while (*p1!='\\'&&p1>=lastfn)
				p1--;

			while (*p2++=*p1++)
				;
#endif
		}
	}
	else // some other file
	{
		if (length_in_ms) // calculate length
		{
			const char *fn=(const char *)fileinfo.buf;

	        memset(&ib, 0, sizeof ib);
			get_plugin()->create_info(&ib, fn);

			if (NULL != ib.track)
			{
				*length_in_ms = ib.length;
				ib.plugin->close(ib.track);
			}
			else *length_in_ms=-1000; // the default is unknown file length (-1000).
		}
		if (title) // get non path portion of filename
		{
#ifdef UNICODE_INPUT_PLUGIN
			const in_char *p=wfilename+wcslen(wfilename);
			while (*p != '\\' && p >= wfilename) p--;
			wcscpy(title,++p);
#else
			const in_char *p=filename+strlen(filename);
			while (*p != '\\' && p >= filename) p--;
			strcpy(title,++p);
#endif
		}
	}

    LeaveCriticalSection(&cs);
}

void eq_set(int on, char data[10], int preamp) 
{ 
	// most plug-ins can't even do an EQ anyhow.. I'm working on writing
	// a generic PCM EQ, but it looks like it'll be a little too CPU 
	// consuming to be useful :)
	// if you _CAN_ do EQ with your format, each data byte is 0-63 (+20db <-> -20db)
	// and preamp is the same. 
}


// render 576 samples into buf. 
// this function is only used by DecodeThread. 

// note that if you adjust the size of buf.samples, for say, 1024
// sample blocks, it will still work, but some of the visualization 
// might not look as good as it could. Stick with 576 sample blocks
// if you can, and have an additional auxiliary (overflow) buffer if 
// necessary.. 
int get_576_samples(char *nouse)
{
    int nsamples;

	//EnterCriticalSection(&cs);
	nsamples = samplebuf_read(&pb);
	//LeaveCriticalSection(&cs);

	return nsamples;
}



DWORD WINAPI DecodeThread(LPVOID b)
{
	int done=0; // set to TRUE if decoding has finished
	pb.plugin->start(pb.track, vout_get(mod.hMainWindow), getoutputtime, settime, showvis);

	while (!killDecodeThread) 
	{
		if (seek_needed != -1) // seek is needed.
		{
#ifdef __SEEK__
			int offs;
			decode_pos_ms = seek_needed;
			seek_needed=-1;
			done=0;
			mod.outMod->Flush(decode_pos_ms); // flush output device and set 
											  // output position to the seek position
			offs = MulDiv(decode_pos_ms,pb.rate,1000); // decode_pos_ms*pb.rate/1000
			SetFilePointer(input_file,offs*pb.nch*(pb.bps/8),NULL,FILE_BEGIN); // seek!
#else
			if (pb.plugin->seek(pb.track, seek_needed)) {
				decode_pos_ms = seek_needed;
				mod.outMod->Flush(decode_pos_ms);
			    pb.plugin->flush(pb.track);
				pb.plugin->restart(pb.track);
			}

			seek_needed=-1;
			done=0;
#endif
		}

		if (done) // done was set to TRUE during decoding, signaling eof
		{
			mod.outMod->CanWrite();		// some output drivers need CanWrite
									    // to be called on a regular basis.

			if (!mod.outMod->IsPlaying()) 
			{
				// we're done playing, so tell Winamp and quit the thread.
				killDecodeThread=1;
				PostMessage(mod.hMainWindow,WM_WA_MPEG_EOF,0,0);
				//return 0;	// quit thread
				//break;	// quit thread
			}
			else
			Sleep(10);		// give a little CPU time back to the system.
		}
		else if (mod.outMod->CanWrite() >= ((pb.nsam*pb.nch*(pb.bps/8))*(mod.dsp_isactive()?2:1)))
			// CanWrite() returns the number of bytes you can write, so we check that
			// to the block size. the reason we multiply the block size by two if 
			// mod.dsp_isactive() is that DSP plug-ins can change it by up to a 
			// factor of two (for tempo adjustment).
		{	
			int l=pb.nsam*pb.nch*(pb.bps/8);			       // block length in bytes
												   // sample buffer. twice as 
												   // big as the blocksize

			l=get_576_samples(pb.samples);	   // retrieve samples

			if (!l)			// no samples means we're at eof
			{
				done=1;
			}
			else	// we got samples!
			{
				// give the samples to the vis subsystems
				mod.SAAddPCMData((char *)pb.samples,pb.nch,pb.bps,decode_pos_ms);	
				mod.VSAAddPCMData((char *)pb.samples,pb.nch,pb.bps,decode_pos_ms);
				// adjust decode position variable
				decode_pos_ms+=(pb.nsam*1000)/pb.rate;	

				// if we have a DSP plug-in, then call it on our samples
				if (mod.dsp_isactive()) 
					l=mod.dsp_dosamples(
						(short *)pb.samples,l/pb.nch/(pb.bps/8),pb.bps,pb.nch,pb.rate
					  ) // dsp_dosamples
					  *(pb.nch*(pb.bps/8));

				// write the pcm data to the output system
				mod.outMod->Write(pb.samples,l);
			}
		}
		else Sleep(20); 
		// if we can't write data, wait a little bit. Otherwise, continue 
		// through the loop writing more data (without sleeping)
	}
	pb.plugin->stop(pb.track);
	return 0;
}


// module definition.

In_Module mod = 
{
	IN_VER,	// defined in IN2.H
	FFSOX " "
	// winamp runs on both alpha systems and x86 ones. :)
#ifdef __alpha
	"(AXP)"
#else
	"(x86)"
#endif
	,
	0,	// hMainWindow (filled in by winamp)
	0,  // hDllInstance (filled in by winamp)
	exts
	// this is a double-null limited list. "EXT\0Description\0EXT\0Description\0" etc.
	,
	1,	// is_seekable
	1,	// uses output plug-in system
	config,
	about,
	init,
	quit,
	getfileinfo,
	infoDlg,
	isourfile,
	play,
	pause,
	unpause,
	ispaused,
	stop,
	
	getlength,
	getoutputtime,
	setoutputtime,

	setvolume,
	setpan,

	0,0,0,0,0,0,0,0,0, // visualization calls filled in by winamp

	0,0, // dsp calls filled in by winamp

	eq_set,

	NULL,		// setinfo call filled in by winamp

	0 // out_mod filled in by winamp

};

// exported symbol. Returns output module.

__declspec( dllexport ) In_Module * winampGetInModule2()
{
	return &mod;
}

#define UNICODE_METADATA
#ifdef UNICODE_METADATA
__declspec (dllexport) int winampGetExtendedFileInfoW (wchar_t *wfilename,
    char *metadata, wchar_t *ret, int retlen)
{
  int code=1;
  samplebuf_t *buf=NULL;

  EnterCriticalSection(&cs);
  strbuf_wchar2utf8(&fileinfo,ini.plugin.path_page,wfilename);

  if (*fileinfo.buf) {
    if (0!=strcmp(ib.fn, (char *)fileinfo.buf)) {
	  const char *fn=(const char *)fileinfo.buf;

	  memset(&ib, 0, sizeof ib);
      get_plugin()->create_info(&ib, fn);

      if (NULL==ib.track)
        goto fail;
      else
        ib.plugin->close(ib.track);
	}

    buf=&ib;
  }
  else if (*pb.fn&&0<strlen(pb.fn))
    buf=&pb;
  else
    goto fail;

  if (0==_stricmp(metadata, "length")) {
	snwprintf(ret, retlen, L"%d", buf->length);
#ifdef __DEBUG_TAGS__
    fprintf(stderr, "+++ \"length\": %d\n", buf->length);
#endif
  }
  else if (0==_stricmp(metadata, "formatinformation")) {
    wstrbuf_utf82wchar(&tag,CP_UTF8,buf->formatinformation);
	snwprintf(ret, retlen, L"%s", tag.buf);
#ifdef __DEBUG_TAGS__
    fprintf(stderr, "+++ \"formatinformation\"\n%s", buf->formatinformation);
#endif
  }
#ifdef __XXX__
  else if (0==_stricmp(metadata, "streamtype"))
	snwprintf(ret, retlen, L"%s", L"streamtype");
  else if (0==_stricmp(metadata, "streamname"))
	snwprintf(ret, retlen, L"%s", L"streamname");
  else if (0==_stricmp(metadata, "GracenoteFileID"))
	snwprintf(ret, retlen, L"%s", L"GracenoteFileID");
  else if (0==_stricmp(metadata, "GracenoteExtData"))
	snwprintf(ret, retlen, L"%s", L"GracenoteExtData");
#endif // __XXX__
#ifdef __DEBUG_TAGS__
#define MAKE_INFO(id) \
  else if (0==_stricmp(#id, metadata)) { \
	snwprintf (ret, retlen, L"%s", buf->tag.id); \
    fprintf(stderr, "+++ \"%s\": \"%s\"\n", #id, buf->tag.id); \
  }
#else // ! __DEBUG_TAGS__
#define MAKE_INFO(id) \
  else if (0==_stricmp(#id, metadata)) { \
    wstrbuf_utf82wchar(&tag,ini.plugin.tag_page,buf->tag.id); \
	snwprintf (ret, retlen, L"%s", tag.buf); \
  }
#endif // __DEBUG_TAGS__
#define MAKE_NEW_INFO(id)
#define MAKE_ALTERNATE_INFO(id1, id2)
#include "info.h"
#undef MAKE_ALTERNATE_INFO
#undef MAKE_NEW_INFO
#undef MAKE_INFO
   else {
#ifdef __DEBUG_TAGS__
     fprintf(stderr, "--- \"%s\"\n", metadata);
#endif
     code=0;
   }

  LeaveCriticalSection(&cs);

  return code;
fail:
#ifdef __DEBUG_TAGS__
  fprintf(stderr, "--- FAIL: \"%s\": \"%s\"\n", metadata, filename);
#endif
  if (0==_stricmp(metadata, "length"))
	snwprintf (ret, retlen, L"%d", 0);
  else
    code=0;

  LeaveCriticalSection(&cs);

  return code;
}
#endif // UNICODE_METADATA

//#define __DEBUG_TAGS__
// http://www.myplugins.info/winamp-wiki/doku.php/advanced_title_formatting
__declspec (dllexport) int winampGetExtendedFileInfo(char *filename,
		char *metadata, char *ret, int retlen)
{
  int code=1;
  samplebuf_t *buf=NULL;

  EnterCriticalSection(&cs);

  if (filename&&*filename) {
    if (0!=strcmp(ib.fn, filename)) {
	  const char *fn=filename;

	  memset(&ib, 0, sizeof ib);
      get_plugin()->create_info(&ib, fn);

      if (NULL==ib.track)
        goto fail;
      else
        ib.plugin->close(ib.track);
	}

    buf=&ib;
  }
  else if (*pb.fn&&0<strlen(pb.fn))
    buf=&pb;
  else
    goto fail;

  if (0==_stricmp(metadata, "length")) {
	_snprintf(ret, retlen, "%d", buf->length);
#ifdef __DEBUG_TAGS__
    fprintf(stderr, "+++ \"length\": %d\n", buf->length);
#endif
  }
  else if (0==_stricmp(metadata, "formatinformation")) {
	_snprintf(ret, retlen, "%s", buf->formatinformation);
#ifdef __DEBUG_TAGS__
    fprintf(stderr, "+++ \"formatinformation\"\n%s", buf->formatinformation);
#endif
  }
#ifdef __XXX__
  else if (0==_stricmp(metadata, "streamtype"))
	_snprintf(ret, retlen, "%s", "streamtype");
  else if (0==_stricmp(metadata, "streamname"))
	_snprintf(ret, retlen, "%s", "streamname");
  else if (0==_stricmp(metadata, "GracenoteFileID"))
	_snprintf(ret, retlen, "%s", "GracenoteFileID");
  else if (0==_stricmp(metadata, "GracenoteExtData"))
	_snprintf(ret, retlen, "%s", "GracenoteExtData");
#endif // __XXX__
#ifdef __DEBUG_TAGS__
#define MAKE_INFO(id) \
  else if (0==_stricmp(#id, metadata)) { \
	_snprintf (ret, retlen, "%s", buf->tag.id); \
    fprintf(stderr, "+++ \"%s\": \"%s\"\n", #id, buf->tag.id); \
  }
#else // ! __DEBUG_TAGS__
#define MAKE_INFO(id) \
  else if (0==_stricmp(#id, metadata)) \
	_snprintf (ret, retlen, "%s", buf->tag.id);
#endif // __DEBUG_TAGS__
#define MAKE_NEW_INFO(id)
#define MAKE_ALTERNATE_INFO(id1, id2)
#include "info.h"
#undef MAKE_ALTERNATE_INFO
#undef MAKE_NEW_INFO
#undef MAKE_INFO
   else {
#ifdef __DEBUG_TAGS__
     fprintf(stderr, "--- \"%s\"\n", metadata);
#endif
     code=0;
   }

  LeaveCriticalSection(&cs);

  return code;
fail:
#ifdef __DEBUG_TAGS__
  fprintf(stderr, "--- FAIL: \"%s\": \"%s\"\n", metadata, filename);
#endif
  if (0==_stricmp(metadata, "length"))
	_snprintf (ret, retlen, "%d", 0);
  else
    code=0;

  LeaveCriticalSection(&cs);

  return code;
}

// return 1 if you want winamp to show it's own file info dialogue, 0 if you want to show your own (via In_Module.InfoBox)
// if returning 1, remember to implement winampGetExtendedFileInfo("formatinformation")!
__declspec(dllexport) int winampUseUnifiedFileInfoDlg(const wchar_t * fn)
{
    return 1;
}

DWORD WINAPI FullScreenThread(LPVOID b)
{
	if (! SendMessage(mod.hMainWindow,WM_WA_IPC,0,IPC_IS_FULLSCREEN)) {
	Sleep(5000);
	// http://cboard.cprogramming.com/cplusplus-programming/15232-console-app-full-screen.html
    keybd_event(VK_MENU,0x38,0,0);
    keybd_event(VK_RETURN,0x1c,0,0);
    keybd_event(VK_RETURN,0x1c,KEYEVENTF_KEYUP,0);
    keybd_event(VK_MENU,0x38,KEYEVENTF_KEYUP,0);
	}

	return 0;
}

void afs()
{
	DWORD thread_id;

	CreateThread(NULL,0,(LPTHREAD_START_ROUTINE) FullScreenThread,NULL,0,&thread_id);
}

void vfs()
{
	if (! SendMessage(mod.hMainWindow,WM_WA_IPC,0,IPC_IS_FULLSCREEN)) {
    keybd_event(VK_MENU,0x38,0,0);
    keybd_event(VK_RETURN,0x1c,0,0);
    keybd_event(VK_RETURN,0x1c,KEYEVENTF_KEYUP,0);
    keybd_event(VK_MENU,0x38,KEYEVENTF_KEYUP,0);
	}
/*
	DWORD thread_id;

	CreateThread(NULL,0,(LPTHREAD_START_ROUTINE) FullScreenThread,NULL,0,&thread_id);
*/
}

void showvis()
{
	int visrunning=SendMessage(mod.hMainWindow,WM_WA_IPC,0,IPC_ISVISRUNNING);

    if (! visrunning)
      PostMessage(mod.hMainWindow,WM_COMMAND,40192,0);
/*
	Sleep(5000);
	// http://cboard.cprogramming.com/cplusplus-programming/15232-console-app-full-screen.html
    keybd_event(VK_MENU,0x38,0,0);
    keybd_event(VK_RETURN,0x1c,0,0);
    keybd_event(VK_RETURN,0x1c,KEYEVENTF_KEYUP,0);
    keybd_event(VK_MENU,0x38,KEYEVENTF_KEYUP,0);
*/
	//afs();
}
