/*
 * plugin.h
 * Copyright (C) 2010 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __PLUGIN_H__
#define __PLUGIN_H__

#ifdef __cplusplus
extern "C" {
#endif

//#define __DEBUG__
//#define __CHOOSE_LIB__

#include <stdio.h>
#include <windows.h>
#include "vout.h"
#define __DYNAMIC__
#include "ffmpeg_stub.h"
#include "libsox_stub.h"

#define FFSOX_PLAYER "FFSoX Player"
#define FFSOX_VERSION "0.4.8"
#define FFSOX FFSOX_PLAYER " v" FFSOX_VERSION

////////////////////////////////////////////////////////////////

#define LIBAVCODEC "libavcodec"
#define LIBA52 "liba52"
#define LIBMAD "libmad"

#define MODULE "in_ffsox.dll"
#define GAIN "gain"
#define VOL "vol"

#define RATE_FIRST  1
#define RATE_LAST   2

typedef struct ini {
  char path[3*MAX_PATH];
  char *p;

  struct {
    char exts[1024];
    int bps;
    int vis;
    int vid;
  } winamp;

  struct {
    int mode;
    double preamp;
    double algorithm_preamp;
	char algorithm[32];
	int comment;
  } rg;

  struct {
    char amplify[16];
    char limit[16];
    int limit_switch;
    int rate_last;
    int sample_rate;
    int largest_multiple;
	char quality[64];
	char phase_response[64];
    int dither;
    int noise_shaping;
  } sox;

  struct {
    int timer_res;
    int sox_decoder;
	int force_seek;
	int prefer_float;

    struct {
      int on;
      int before;
      int after;
    } drop_frames;

    char _ac3_decoder[32];
    char _mp3_decoder[32];
    int tag_page;
    int path_page;
  } plugin;

  struct {
    struct {
      int min;
      int max;
    } audio;

    struct {
      int min;
      int max;
    } video;
  } queues;
} ini_t;

extern ini_t ini;
extern char exts[4096];

void load_ini();

BOOL CALLBACK ConfigDlgProc(HWND hDlg, UINT message, WPARAM wParam,
    LPARAM lParam);
BOOL CALLBACK AboutDlgProc(HWND hDlg, UINT message, WPARAM wParam,
    LPARAM lParam);

////////////////////////////////////////////////////////////////

typedef struct track track_t;
typedef struct plugin plugin_t;

typedef enum samplebuf_state {
  SAMPLEBUF_START,
  SAMPLEBUF_READ,
  SAMPLEBUF_WRITE,
  SAMPLEBUF_DONE
} samplebuf_state_t;

typedef struct samplebuf {
  plugin_t *plugin;
  track_t *track;

  char fn[3*MAX_PATH];
  char csv[3*MAX_PATH];
  char formatinformation[1028];

  int nsam;
  int nch;
  int rate;
  int bps;
  int bytes;
  int length;

  struct tag {
    double gain;
    double q;
#define MAKE_INFO(id) \
    char id[128];
#define MAKE_NEW_INFO(id) MAKE_INFO(id)
#define MAKE_ALTERNATE_INFO(id1, id2)
#include "info.h"
#undef MAKE_ALTERNATE_INFO
#undef MAKE_NEW_INFO
#undef MAKE_INFO
  } tag;

  size_t size;
  char *samples;
  char *mp;
  volatile char *wp;
  volatile samplebuf_state_t state;

  HANDLE hEventRead;
  HANDLE hEventWrite;
} samplebuf_t;

samplebuf_t *samplebuf_open(samplebuf_t *buf, plugin_t *plugin,
    track_t *track, const char *fn, int nch, int rate, int bps,
    int length);
void samplebuf_close(samplebuf_t *buf);

samplebuf_t *samplebuf_start(samplebuf_t *buf);
void samplebuf_stop(samplebuf_t *buf);

char *samplebuf_get_gain_val(samplebuf_t *buf);
double samplebuf_get_gain(samplebuf_t *buf);

void samplebuf_reset(samplebuf_t *buf);
void samplebuf_signal_done(samplebuf_t *buf);
void samplebuf_signal_read(samplebuf_t *buf);
void samplebuf_signal_write(samplebuf_t *buf);

int samplebuf_wait_read(samplebuf_t *buf);
int samplebuf_wait_write(samplebuf_t *buf);

int samplebuf_read(samplebuf_t *buf);

void prefs_set_gain(struct tag *tag);

////////////////////////////////////////////////////////////////

typedef struct timer {
  CRITICAL_SECTION cs;
  HANDLE hEvent;
  UINT wTimerRes;
  volatile MMRESULT wTimerID;
} timer_t;

timer_t *timer_open(timer_t *t, UINT uPeriod);
void timer_close(timer_t *t);

void timer_kill(timer_t *t);
void timer_sleep(timer_t *t, UINT msInterval);

////////////////////////////////////////////////////////////////

sox_effects_chain_t *track_init_sox_chain(track_t *track,
    samplebuf_t *buf, int in_bps, int in_rate, int in_channels,
	int apply_gain, sox_effect_handler_t const *(*input_handler)(void));

////////////////////////////////////////////////////////////////

typedef enum packet_queue_state {
  PACKET_QUEUE_WORK,
  PACKET_QUEUE_PAUSE,
  PACKET_QUEUE_KILL,
  PACKET_QUEUE_EOF
} packet_queue_state_t;

typedef struct packet_queue {
  volatile AVPacketList *head;
  volatile AVPacketList *tail;

  int min;
  int max;
  volatile int cur;

  volatile packet_queue_state_t state;
  HANDLE hEventWrite;
  HANDLE hEventRead;
  CRITICAL_SECTION cs;
} packet_queue_t;

packet_queue_t *packet_queue_open(packet_queue_t *q, int min, int max,
    HANDLE hEventWrite);
void packet_queue_close(packet_queue_t *q);

void packet_queue_reset(packet_queue_t *q);
void packet_queue_signal_work(packet_queue_t *q);
int packet_queue_overflow(packet_queue_t *q);
int packet_queue_underflow(packet_queue_t *q);
void packet_queue_write(packet_queue_t *q, AVPacket *pkt);
AVPacket *packet_queue_read(packet_queue_t *q, AVPacket *pkt);

////////////////////////////////////////////////////////////////

struct plugin {
  track_t *(*create_play)(samplebuf_t *buf, const char *fn);
  track_t *(*create_info)(samplebuf_t *buf, const char *fn);
  void (*close)(track_t *t);
  track_t *(*start)(track_t *t, vout_t *vout, int (*gettime)(),
      void (*settime)(int), void (*showvis)());
  void (*stop)(track_t *t);
  void (*kill)(track_t *t);
  void (*pause)(track_t *t);
  void (*unpause)(track_t *t);
  int (*seek)(track_t *track, int ms);
  void (*flush)(track_t *t);
  void (*restart)(track_t *t);
};

extern plugin_t sox_plugin;
extern plugin_t ffsox_plugin;

#ifdef __cplusplus
}
#endif

#endif // __PLUGIN_H__
