/*
 * prefs.c
 * Copyright (C) 2010 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "resources.h"
#include "plugin.h"
#include <commctrl.h>

//#define DEF_EXTS "AVI;MKV;MOV;MP3;MP4;RM;VOB;FLAC;FLV;WV;OGG"
#define DEF_EXTS "FLAC;MKV;OGG;WAV;WEBM;WV"
#define AC3_DECODER "ac3_decoder"
#define MP3_DECODER "mp3_decoder"

#define PREAMP_MIN (-12.0)
#define PREAMP_MAX (12.0)
#define PREAMP_SCALE (2.0)

#define PACKET_QUEUE_MIN 4
#define PACKET_QUEUE_MAX 8

#define PICTURE_QUEUE_MIN 4
#define PICTURE_QUEUE_MAX 8

#define LOCATION "in_ffsox\\in_ffsox.ini"

ini_t ini;
char exts[4096];
static char tmp_rg_algorithm[32];
static double tmp_rg_preamp;
static double tmp_rg_algorithm_preamp;

typedef struct rate_last_prefs {
  const char *label;
  const int rate_last;
} rate_last_prefs_t;

typedef struct resample_prefs {
  const char *label;
  const int rate;
} resample_prefs_t;

typedef struct quality_prefs {
  const char *label;
  const char *quality;
} quality_prefs_t;

typedef struct phase_response_prefs {
  const char *label;
  const char *phase_response;
} phase_response_prefs_t;

typedef struct limit_gain_prefs {
  const char *label;
  const char *gain;
} limit_gain_prefs_t;

typedef struct rg_algorithm_prefs {
  const char *label;
  const char *algorithm;
  const char *label2;
} rg_algorithm_prefs_t;

typedef struct page {
  int page;
  const char *label;
} page_t;

static rate_last_prefs_t rate_last_prefs[] = {
  { "default", 0 },
  { "first", RATE_FIRST },
  { "last", RATE_LAST },
  { NULL, 0 }
};

static resample_prefs_t resample_prefs[] = {
  { "off", 0 },
  { "11025", 11025 },
  { "22050", 22050 },
  { "24000", 24000 },
  { "32000", 32000 },
  { "44100", 44100 },
  { "48000", 48000 },
  { "88200", 88200 },
  { "96000", 96000 },
  { "176400", 176400 },
  { "192000", 192000 },
  { NULL, 0 }
};

static quality_prefs_t quality_prefs[] = {
  { "default", "" },
  { "quick", "-q" },
  { "low", "-l" },
  { "medium", "-m" },
  { "high", "-h" },
  { "very high", "-v" },
  { NULL, NULL }
};

static phase_response_prefs_t phase_response_prefs[] = {
  { "default", "" },
  { "minimum", "-M" },
  { "intermediate", "-I" },
  { "linear", "-L" },
  { NULL, 0 }
};

static limit_gain_prefs_t limit_gain_prefs[] = {
  { "0.01", "0.01" },
  { "0.02", "0.02" },
  { "0.03", "0.03" },
  { "0.04", "0.04" },
  { "0.05", "0.05" },
  { "0.06", "0.06" },
  { "0.07", "0.07" },
  { "0.08", "0.08" },
  { "0.09", "0.09" },
  { NULL, NULL }
};

static rg_algorithm_prefs_t rg_algorithm_prefs[] = {
  //{ "ReplayGain", "rg", "EBU R128 Preamp:" },
  //{ "EBU R128", "r128" , "ReplayGain Preamp:"},
  { "ReplayGain", "rg", "EBU R128:" },
  { "EBU R128", "r128" , "ReplayGain:"},
  { NULL, NULL, NULL }
};

static page_t pages[]={
// http://msdn.microsoft.com/en-us/library/dd317756%28v=VS.85%29.aspx
{ 37, "IBM EBCDIC US-Canada" },
{ 437, "OEM United States" },
{ 500, "IBM EBCDIC International" },
{ 708, "Arabic (ASMO 708)" },
{ 709, "Arabic (ASMO-449+, BCON V4)" },
{ 710, "Arabic - Transparent Arabic" },
{ 720, "Arabic (Transparent ASMO); Arabic (DOS)" },
{ 737, "OEM Greek (formerly 437G); Greek (DOS)" },
{ 775, "OEM Baltic; Baltic (DOS)" },
{ 850, "OEM Multilingual Latin 1; Western European (DOS)" },
{ 852, "OEM Latin 2; Central European (DOS)" },
{ 855, "OEM Cyrillic (primarily Russian)" },
{ 857, "OEM Turkish; Turkish (DOS)" },
{ 858, "OEM Multilingual Latin 1 + Euro symbol" },
{ 860, "OEM Portuguese; Portuguese (DOS)" },
{ 861, "OEM Icelandic; Icelandic (DOS)" },
{ 862, "OEM Hebrew; Hebrew (DOS)" },
{ 863, "OEM French Canadian; French Canadian (DOS)" },
{ 864, "OEM Arabic; Arabic (864)" },
{ 865, "OEM Nordic; Nordic (DOS)" },
{ 866, "OEM Russian; Cyrillic (DOS)" },
{ 869, "OEM Modern Greek; Greek, Modern (DOS)" },
{ 870, "IBM EBCDIC Multilingual/ROECE (Latin 2); IBM EBCDIC Multilingual Latin 2" },
{ 874, "ANSI/OEM Thai (same as 28605, ISO 8859-15); Thai (Windows)" },
{ 875, "IBM EBCDIC Greek Modern" },
{ 932, "ANSI/OEM Japanese; Japanese (Shift-JIS)" },
{ 936, "ANSI/OEM Simplified Chinese (PRC, Singapore); Chinese Simplified (GB2312)" },
{ 949, "ANSI/OEM Korean (Unified Hangul Code)" },
{ 950, "ANSI/OEM Traditional Chinese (Taiwan; Hong Kong SAR, PRC); Chinese Traditional (Big5)" },
{ 1026, "IBM EBCDIC Turkish (Latin 5)" },
{ 1047, "IBM EBCDIC Latin 1/Open System" },
{ 1140, "IBM EBCDIC US-Canada (037 + Euro symbol); IBM EBCDIC (US-Canada-Euro)" },
{ 1141, "IBM EBCDIC Germany (20273 + Euro symbol); IBM EBCDIC (Germany-Euro)" },
{ 1142, "IBM EBCDIC Denmark-Norway (20277 + Euro symbol); IBM EBCDIC (Denmark-Norway-Euro)" },
{ 1143, "IBM EBCDIC Finland-Sweden (20278 + Euro symbol); IBM EBCDIC (Finland-Sweden-Euro)" },
{ 1144, "IBM EBCDIC Italy (20280 + Euro symbol); IBM EBCDIC (Italy-Euro)" },
{ 1145, "IBM EBCDIC Latin America-Spain (20284 + Euro symbol); IBM EBCDIC (Spain-Euro)" },
{ 1146, "IBM EBCDIC United Kingdom (20285 + Euro symbol); IBM EBCDIC (UK-Euro)" },
{ 1147, "IBM EBCDIC France (20297 + Euro symbol); IBM EBCDIC (France-Euro)" },
{ 1148, "IBM EBCDIC International (500 + Euro symbol); IBM EBCDIC (International-Euro)" },
{ 1149, "IBM EBCDIC Icelandic (20871 + Euro symbol); IBM EBCDIC (Icelandic-Euro)" },
{ 1200, "Unicode UTF-16, little endian byte order (BMP of ISO 10646); available only to managed applications" },
{ 1201, "Unicode UTF-16, big endian byte order; available only to managed applications" },
{ 1250, "ANSI Central European; Central European (Windows)" },
{ 1251, "ANSI Cyrillic; Cyrillic (Windows)" },
{ 1252, "ANSI Latin 1; Western European (Windows)" },
{ 1253, "ANSI Greek; Greek (Windows)" },
{ 1254, "ANSI Turkish; Turkish (Windows)" },
{ 1255, "ANSI Hebrew; Hebrew (Windows)" },
{ 1256, "ANSI Arabic; Arabic (Windows)" },
{ 1257, "ANSI Baltic; Baltic (Windows)" },
{ 1258, "ANSI/OEM Vietnamese; Vietnamese (Windows)" },
{ 1361, "Korean (Johab)" },
{ 10000, "MAC Roman; Western European (Mac)" },
{ 10001, "Japanese (Mac)" },
{ 10002, "MAC Traditional Chinese (Big5); Chinese Traditional (Mac)" },
{ 10003, "Korean (Mac)" },
{ 10004, "Arabic (Mac)" },
{ 10005, "Hebrew (Mac)" },
{ 10006, "Greek (Mac)" },
{ 10007, "Cyrillic (Mac)" },
{ 10008, "MAC Simplified Chinese (GB 2312); Chinese Simplified (Mac)" },
{ 10010, "Romanian (Mac)" },
{ 10017, "Ukrainian (Mac)" },
{ 10021, "Thai (Mac)" },
{ 10029, "MAC Latin 2; Central European (Mac)" },
{ 10079, "Icelandic (Mac)" },
{ 10081, "Turkish (Mac)" },
{ 10082, "Croatian (Mac)" },
{ 12000, "Unicode UTF-32, little endian byte order; available only to managed applications" },
{ 12001, "Unicode UTF-32, big endian byte order; available only to managed applications" },
{ 20000, "CNS Taiwan; Chinese Traditional (CNS)" },
{ 20001, "TCA Taiwan" },
{ 20002, "Eten Taiwan; Chinese Traditional (Eten)" },
{ 20003, "IBM5550 Taiwan" },
{ 20004, "TeleText Taiwan" },
{ 20005, "Wang Taiwan" },
{ 20105, "IA5 (IRV International Alphabet No. 5, 7-bit); Western European (IA5)" },
{ 20106, "IA5 German (7-bit)" },
{ 20107, "IA5 Swedish (7-bit)" },
{ 20108, "IA5 Norwegian (7-bit)" },
{ 20127, "US-ASCII (7-bit)" },
{ 20261, "T.61" },
{ 20269, "ISO 6937 Non-Spacing Accent" },
{ 20273, "IBM EBCDIC Germany" },
{ 20277, "IBM EBCDIC Denmark-Norway" },
{ 20278, "IBM EBCDIC Finland-Sweden" },
{ 20280, "IBM EBCDIC Italy" },
{ 20284, "IBM EBCDIC Latin America-Spain" },
{ 20285, "IBM EBCDIC United Kingdom" },
{ 20290, "IBM EBCDIC Japanese Katakana Extended" },
{ 20297, "IBM EBCDIC France" },
{ 20420, "IBM EBCDIC Arabic" },
{ 20423, "IBM EBCDIC Greek" },
{ 20424, "IBM EBCDIC Hebrew" },
{ 20833, "IBM EBCDIC Korean Extended" },
{ 20838, "IBM EBCDIC Thai" },
{ 20866, "Russian (KOI8-R); Cyrillic (KOI8-R)" },
{ 20871, "IBM EBCDIC Icelandic" },
{ 20880, "IBM EBCDIC Cyrillic Russian" },
{ 20905, "IBM EBCDIC Turkish" },
{ 20924, "IBM EBCDIC Latin 1/Open System (1047 + Euro symbol)" },
{ 20932, "Japanese (JIS 0208-1990 and 0121-1990)" },
{ 20936, "Simplified Chinese (GB2312); Chinese Simplified (GB2312-80)" },
{ 20949, "Korean Wansung" },
{ 21025, "IBM EBCDIC Cyrillic Serbian-Bulgarian" },
{ 21027, "(deprecated)" },
{ 21866, "Ukrainian (KOI8-U); Cyrillic (KOI8-U)" },
{ 28591, "ISO 8859-1 Latin 1; Western European (ISO)" },
{ 28592, "ISO 8859-2 Central European; Central European (ISO)" },
{ 28593, "ISO 8859-3 Latin 3" },
{ 28594, "ISO 8859-4 Baltic" },
{ 28595, "ISO 8859-5 Cyrillic" },
{ 28596, "ISO 8859-6 Arabic" },
{ 28597, "ISO 8859-7 Greek" },
{ 28598, "ISO 8859-8 Hebrew; Hebrew (ISO-Visual)" },
{ 28599, "ISO 8859-9 Turkish" },
{ 28603, "ISO 8859-13 Estonian" },
{ 28605, "ISO 8859-15 Latin 9" },
{ 29001, "Europa 3" },
{ 38598, "ISO 8859-8 Hebrew; Hebrew (ISO-Logical)" },
{ 50220, "ISO 2022 Japanese with no halfwidth Katakana; Japanese (JIS)" },
{ 50221, "ISO 2022 Japanese with halfwidth Katakana; Japanese (JIS-Allow 1 byte Kana)" },
{ 50222, "ISO 2022 Japanese JIS X 0201-1989; Japanese (JIS-Allow 1 byte Kana - SO/SI)" },
{ 50225, "ISO 2022 Korean" },
{ 50227, "ISO 2022 Simplified Chinese; Chinese Simplified (ISO 2022)" },
{ 50229, "ISO 2022 Traditional Chinese" },
{ 50930, "EBCDIC Japanese (Katakana) Extended" },
{ 50931, "EBCDIC US-Canada and Japanese" },
{ 50933, "EBCDIC Korean Extended and Korean" },
{ 50935, "EBCDIC Simplified Chinese Extended and Simplified Chinese" },
{ 50936, "EBCDIC Simplified Chinese" },
{ 50937, "EBCDIC US-Canada and Traditional Chinese" },
{ 50939, "EBCDIC Japanese (Latin) Extended and Japanese" },
{ 51932, "EUC Japanese" },
{ 51936, "EUC Simplified Chinese; Chinese Simplified (EUC)" },
{ 51949, "EUC Korean" },
{ 51950, "EUC Traditional Chinese" },
{ 52936, "HZ-GB2312 Simplified Chinese; Chinese Simplified (HZ)" },
{ 54936, "Windows XP and later: GB18030 Simplified Chinese (4 byte); Chinese Simplified (GB18030)" },
{ 57002, "ISCII Devanagari" },
{ 57003, "ISCII Bengali" },
{ 57004, "ISCII Tamil" },
{ 57005, "ISCII Telugu" },
{ 57006, "ISCII Assamese" },
{ 57007, "ISCII Oriya" },
{ 57008, "ISCII Kannada" },
{ 57009, "ISCII Malayalam" },
{ 57010, "ISCII Gujarati" },
{ 57011, "ISCII Punjabi" },
{ 65000, "Unicode (UTF-7)" },
{ 65001, "Unicode (UTF-8)" },
{ 0, NULL }
};

static void get_exts()
{
  char *rp=ini.winamp.exts;
  char *mrp=ini.winamp.exts+strlen(ini.winamp.exts);
  char *wp=exts;
  char *mwp=exts+sizeof(exts)-4;

  memset(exts, 0, sizeof exts);

  while (rp<mrp&&wp<mwp) {
    if (';'==*rp) {
      *wp++=0;
      *wp++='X';
      *wp++=0;
      *wp=0;
      ++rp;
    }
    else
      *wp++=*rp++;
  }
}

static void fill_ini(const char *path, char *p) { 
    char buf[32];

    strcpy(p, LOCATION);

    // winamp
    GetPrivateProfileString("winamp", "exts", DEF_EXTS,
            ini.winamp.exts, sizeof ini.winamp.exts,
            ini.path);
    get_exts();
    ini.winamp.bps=GetPrivateProfileInt("winamp", "bps",
            16, ini.path);
    ini.winamp.vis=GetPrivateProfileInt("winamp", "vis",
            1, ini.path);
    ini.winamp.vid=GetPrivateProfileInt("winamp", "vid",
            1, ini.path);

    // rg
    ini.rg.mode=GetPrivateProfileInt("rg", "mode", 1, ini.path);
    GetPrivateProfileString("rg", "preamp", "0",
            buf, sizeof buf, ini.path);
    tmp_rg_preamp=ini.rg.preamp=atof(buf);
    GetPrivateProfileString("rg", "algorithm", "r128",
            ini.rg.algorithm, sizeof ini.rg.algorithm, ini.path);
	strcpy(tmp_rg_algorithm, ini.rg.algorithm);
    GetPrivateProfileString("rg", "algorithm_preamp", "-5.0",
            buf, sizeof buf, ini.path);
    tmp_rg_algorithm_preamp=ini.rg.algorithm_preamp=atof(buf);
    ini.rg.comment=GetPrivateProfileInt("rg", "comment", 1, ini.path);

    // sox
    GetPrivateProfileString("sox", "amplify", GAIN,
            ini.sox.amplify,
            sizeof ini.sox.amplify,
            ini.path);
    ini.sox.limit_switch=GetPrivateProfileInt("sox", "limit_switch",
            1, ini.path);
    GetPrivateProfileString("sox", "limit", "0.05",
            ini.sox.limit, sizeof ini.sox.limit,
            ini.path);
    ini.sox.sample_rate=GetPrivateProfileInt("sox", "sample_rate",
            0, ini.path);
    ini.sox.rate_last=GetPrivateProfileInt("sox", "rate_last",
            0, ini.path);
    GetPrivateProfileString("sox", "quality", "", ini.sox.quality,
			sizeof ini.sox.quality, ini.path);
    GetPrivateProfileString("sox", "phase_response", "",
			ini.sox.phase_response, sizeof ini.sox.phase_response, ini.path);
    ini.sox.largest_multiple=GetPrivateProfileInt("sox", "largest_multiple",
			0, ini.path);
    ini.sox.dither=GetPrivateProfileInt("sox", "dither",
            1, ini.path);
    ini.sox.noise_shaping=GetPrivateProfileInt("sox", "noise_shaping",
            1, ini.path);

    // plugin
    ini.plugin.timer_res=GetPrivateProfileInt("plugin",
            "timer_res", 4, ini.path);
#ifndef __SOX_DECODER__
    ini.plugin.sox_decoder=0;
#else // __SOX_DECODER__
    ini.plugin.sox_decoder=GetPrivateProfileInt("plugin",
            "sox_decoder", 0, ini.path);
#endif // __SOX_DECODER__
    ini.plugin.drop_frames.on=GetPrivateProfileInt("plugin",
            "drop_frames", 1, ini.path);
    ini.plugin.drop_frames.before=GetPrivateProfileInt("plugin",
            "drop_frames_before", -1000, ini.path);
    ini.plugin.drop_frames.after=GetPrivateProfileInt("plugin",
            "drop_frames_after", 1000, ini.path);

#ifdef __CHOOSE_LIB__
    GetPrivateProfileString("plugin", AC3_DECODER, LIBA52,
            ini.plugin._ac3_decoder,
            sizeof ini.plugin._ac3_decoder,
            ini.path);
    GetPrivateProfileString("plugin", MP3_DECODER, LIBMAD,
            ini.plugin._mp3_decoder,
            sizeof ini.plugin._mp3_decoder,
            ini.path);
#else
	strcpy(ini.plugin._mp3_decoder,LIBAVCODEC);
	strcpy(ini.plugin._ac3_decoder,LIBAVCODEC);
#endif

    ini.plugin.force_seek=GetPrivateProfileInt("plugin", "force_seek",
            0, ini.path);
    ini.plugin.prefer_float=GetPrivateProfileInt("plugin", "prefer_float",
            1, ini.path);
    ini.plugin.tag_page=GetPrivateProfileInt("plugin", "tag_page",
            CP_UTF8, ini.path);
    ini.plugin.path_page=GetPrivateProfileInt("plugin", "path_page",
            CP_UTF8, ini.path);

    // queues
    ini.queues.audio.min=GetPrivateProfileInt("queues",
            "audio_min", PACKET_QUEUE_MIN, ini.path);
    ini.queues.audio.max=GetPrivateProfileInt("queues",
            "audio_max", PACKET_QUEUE_MAX, ini.path);
    ini.queues.video.min=GetPrivateProfileInt("queues",
            "video_min", PACKET_QUEUE_MIN, ini.path);
    ini.queues.video.max=GetPrivateProfileInt("queues",
            "video_max", PACKET_QUEUE_MAX, ini.path);
}

void load_ini()
{
  HMODULE module = GetModuleHandle (MODULE);

  if (!module||!GetModuleFileName(module, ini.path, sizeof ini.path))
    return;

  ini.p=ini.path+strlen(ini.path);

  while (ini.path<ini.p&&'\\'!=ini.p[-1])
    --ini.p;

  fill_ini(ini.path, ini.p);
}

static void save_ini()
{
    char buf[32];

    strcpy(ini.p, LOCATION);

    // winamp
    WritePrivateProfileString("winamp", "exts", ini.winamp.exts,
            ini.path);
    sprintf(buf, "%d", ini.winamp.bps);
    WritePrivateProfileString("winamp", "bps", buf, ini.path);
    sprintf(buf, "%d", ini.winamp.vis);
    WritePrivateProfileString("winamp", "vis", buf, ini.path);
    sprintf(buf, "%d", ini.winamp.vid);
    WritePrivateProfileString("winamp", "vid", buf, ini.path);

    // rg
    sprintf(buf, "%d", ini.rg.mode);
    WritePrivateProfileString("rg", "mode", buf, ini.path);
    sprintf(buf, "%.1f", 0.1*round(10.0*ini.rg.preamp));
    WritePrivateProfileString("rg", "preamp", buf, ini.path);
    sprintf(buf, "%.1f", 0.1*round(10.0*ini.rg.algorithm_preamp));
    WritePrivateProfileString("rg", "algorithm_preamp", buf, ini.path);
    WritePrivateProfileString("rg", "algorithm", ini.rg.algorithm, ini.path);
    sprintf(buf, "%d", ini.rg.comment);
    WritePrivateProfileString("rg", "comment", buf, ini.path);

    // sox
    WritePrivateProfileString("sox", "amplify", ini.sox.amplify, ini.path);
    sprintf(buf, "%d", ini.sox.limit_switch);
    WritePrivateProfileString("sox", "limit_switch", buf, ini.path);
    WritePrivateProfileString("sox", "limit", ini.sox.limit, ini.path);
    sprintf(buf, "%d", ini.sox.sample_rate);
    WritePrivateProfileString("sox", "sample_rate", buf, ini.path);
    sprintf(buf, "%d", ini.sox.rate_last);
    WritePrivateProfileString("sox", "rate_last", buf, ini.path);
    WritePrivateProfileString("sox", "quality", ini.sox.quality, ini.path);
    WritePrivateProfileString("sox", "phase_response", ini.sox.phase_response, ini.path);
    sprintf(buf, "%d", ini.sox.largest_multiple);
    WritePrivateProfileString("sox", "largest_multiple", buf, ini.path);
    sprintf(buf, "%d", ini.sox.dither);
    WritePrivateProfileString("sox", "dither", buf, ini.path);
    sprintf(buf, "%d", ini.sox.noise_shaping);
    WritePrivateProfileString("sox", "noise_shaping", buf, ini.path);

    // plugin
    sprintf(buf, "%d", ini.plugin.timer_res);
    WritePrivateProfileString("plugin", "timer_res", buf, ini.path);
    sprintf(buf, "%d", ini.plugin.sox_decoder);
    WritePrivateProfileString("plugin", "sox_decoder", buf, ini.path);
    sprintf(buf, "%d", ini.plugin.drop_frames.on);
    WritePrivateProfileString("plugin", "drop_frames", buf, ini.path);
    sprintf(buf, "%d", ini.plugin.drop_frames.before);
    WritePrivateProfileString("plugin", "drop_frames_before", buf,
            ini.path);
    sprintf(buf, "%d", ini.plugin.drop_frames.after);
    WritePrivateProfileString("plugin", "drop_frames_after", buf,
            ini.path);
#ifdef __CHOOSE_LIB__
    WritePrivateProfileString("plugin", AC3_DECODER,
            ini.plugin._ac3_decoder, ini.path);
    WritePrivateProfileString("plugin", MP3_DECODER,
            ini.plugin._mp3_decoder, ini.path);
#endif
    sprintf(buf, "%d", ini.plugin.force_seek);
    WritePrivateProfileString("plugin", "force_seek", buf, ini.path);
    sprintf(buf, "%d", ini.plugin.prefer_float);
    WritePrivateProfileString("plugin", "prefer_float", buf, ini.path);
    sprintf(buf, "%d", ini.plugin.tag_page);
    WritePrivateProfileString("plugin", "tag_page", buf, ini.path);
    sprintf(buf, "%d", ini.plugin.path_page);
    WritePrivateProfileString("plugin", "path_page", buf, ini.path);

    // queues
    sprintf(buf, "%d", ini.queues.audio.min);
    WritePrivateProfileString("queues", "audio_min", buf, ini.path);
    sprintf(buf, "%d", ini.queues.audio.max);
    WritePrivateProfileString("queues", "audio_max", buf, ini.path);
    sprintf(buf, "%d", ini.queues.video.min);
    WritePrivateProfileString("queues", "video_min", buf, ini.path);
    sprintf(buf, "%d", ini.queues.video.max);
    WritePrivateProfileString("queues", "video_max", buf, ini.path);
}

static void enable_drop_frames(HWND hDlg, BOOL on)
{
    EnableWindow(GetDlgItem (hDlg, IDC_STATIC_PLUGIN_DROP_FRAMES_BEFORE), on);
    EnableWindow(GetDlgItem (hDlg, IDC_EDIT_PLUGIN_DROP_FRAMES_BEFORE), on);
    EnableWindow(GetDlgItem (hDlg, IDC_STATIC_PLUGIN_DROP_FRAMES_AFTER), on);
    EnableWindow(GetDlgItem (hDlg, IDC_EDIT_PLUGIN_DROP_FRAMES_AFTER), on);
    EnableWindow(GetDlgItem (hDlg, IDC_STATIC_PLUGIN_DROP_FRAMES_MS), on);
}

static void enable_dither(HWND hDlg, BOOL on)
{
    EnableWindow(GetDlgItem (hDlg, IDC_STATIC_SOX_NOISE_SHAPING), on);
    EnableWindow(GetDlgItem (hDlg, IDC_CHECKBOX_SOX_NOISE_SHAPING), on);
}

static void enable_resample(HWND hDlg, BOOL on)
{
    EnableWindow(GetDlgItem (hDlg, IDC_STATIC_SOX_LARGEST_MULTIPLE), on);
    EnableWindow(GetDlgItem (hDlg, IDC_CHECKBOX_SOX_LARGEST_MULTIPLE), on);
    EnableWindow(GetDlgItem (hDlg, IDC_COMBO_SOX_RATE_QUALITY), on);
    EnableWindow(GetDlgItem (hDlg, IDC_COMBO_SOX_RATE_PHASE_RESPONSE), on);
    EnableWindow(GetDlgItem (hDlg, IDC_STATIC_SOX_RATE_PHASE_RESPONSE), on);
    EnableWindow(GetDlgItem (hDlg, IDC_COMBO_SOX_RATE_LAST), on);
    EnableWindow(GetDlgItem (hDlg, IDC_STATIC_SOX_RATE_LAST), on);
}

static void enable_vol(HWND hDlg, BOOL on)
{
    EnableWindow(GetDlgItem (hDlg, IDC_COMBO_SOX_LIMIT_GAIN),
	    on&&IsDlgButtonChecked (hDlg, LOWORD(IDC_CHECKBOX_SOX_LIMIT_SWITCH)));
}

static void enable_rg(HWND hDlg, BOOL on)
{
    EnableWindow(GetDlgItem (hDlg, IDC_STATIC_RG_PREAMP), on);
    EnableWindow(GetDlgItem (hDlg, IDC_STATIC_RG_PREAMP_DB), on);
    EnableWindow(GetDlgItem (hDlg, IDC_SLIDER_RG_PREAMP), on);

    EnableWindow(GetDlgItem (hDlg, IDC_STATIC_RG_ALGORITHM), on);
    EnableWindow(GetDlgItem (hDlg, IDC_COMBO_RG_ALGORITHM), on);

    EnableWindow(GetDlgItem (hDlg, IDC_STATIC_RG_ALGORITHM_PREAMP), on);
    EnableWindow(GetDlgItem (hDlg, IDC_STATIC_RG_ALGORITHM_PREAMP_DB), on);
    EnableWindow(GetDlgItem (hDlg, IDC_SLIDER_RG_ALGORITHM_PREAMP), on);

    EnableWindow(GetDlgItem (hDlg, IDC_STATIC_RG_COMMENT), on);
    EnableWindow(GetDlgItem (hDlg, IDC_CHECKBOX_RG_COMMENT), on);
}

static void select_algorithm(HWND hDlg, int i)
{
  if (0!=strcmp(tmp_rg_algorithm, rg_algorithm_prefs[i].algorithm)) {
    SetDlgItemText(hDlg, IDC_STATIC_RG_ALGORITHM_PREAMP,
      rg_algorithm_prefs[i].label2);

	strcpy(tmp_rg_algorithm, rg_algorithm_prefs[i].algorithm);
	  tmp_rg_preamp+=tmp_rg_algorithm_preamp;
	  tmp_rg_algorithm_preamp=-tmp_rg_algorithm_preamp;

      SendDlgItemMessage(hDlg, IDC_SLIDER_RG_PREAMP, TBM_SETPOS,
          TRUE, PREAMP_SCALE*(tmp_rg_preamp-PREAMP_MIN));

      SendDlgItemMessage(hDlg, IDC_SLIDER_RG_ALGORITHM_PREAMP, TBM_SETPOS,
          TRUE, PREAMP_SCALE*(tmp_rg_algorithm_preamp-PREAMP_MIN));
  }
}

BOOL CALLBACK ConfigDlgProc (HWND hDlg, UINT message, WPARAM wParam,
        LPARAM lParam)
{
  extern int a52_stub_valid;
  extern int mad_stub_valid;

#ifdef __CHOOSE_LIB__
  char *_ac3_decoder=a52_stub_valid?ini.plugin._ac3_decoder:LIBAVCODEC;
  char *_mp3_decoder=mad_stub_valid?ini.plugin._mp3_decoder:LIBAVCODEC;
#else
  char *_ac3_decoder=LIBAVCODEC;
  char *_mp3_decoder=LIBAVCODEC;
#endif
  char buf[32];
  int i, j;

  switch (message) {
    case WM_INITDIALOG:
      // winamp
      SetDlgItemText(hDlg, IDC_EDIT_WINAMP_EXTS, ini.winamp.exts);
      SendDlgItemMessage(hDlg, IDC_COMBO_WINAMP_BPS, CB_ADDSTRING, 0,
          (LPARAM)"16");
      SendDlgItemMessage(hDlg, IDC_COMBO_WINAMP_BPS, CB_ADDSTRING, 0,
          (LPARAM)"24");
      SendDlgItemMessage(hDlg, IDC_COMBO_WINAMP_BPS, CB_SETCURSEL,
          ini.winamp.bps/8-2, 0);
      CheckDlgButton(hDlg, IDC_CHECKBOX_WINAMP_VIS, ini.winamp.vis);
      CheckDlgButton(hDlg, IDC_CHECKBOX_WINAMP_VID, ini.winamp.vid);

      // rg
      SendDlgItemMessage(hDlg, IDC_COMBO_RG_MODE, CB_ADDSTRING, 0,
          (LPARAM)"off");
      SendDlgItemMessage(hDlg, IDC_COMBO_RG_MODE, CB_ADDSTRING, 0,
          (LPARAM)"track");
      SendDlgItemMessage(hDlg, IDC_COMBO_RG_MODE, CB_ADDSTRING, 0,
          (LPARAM)"album");
      SendDlgItemMessage(hDlg, IDC_COMBO_RG_MODE, CB_SETCURSEL,
          ini.rg.mode, 0);
      CheckDlgButton(hDlg, IDC_CHECKBOX_RG_COMMENT, ini.rg.comment);

      SendDlgItemMessage(hDlg, IDC_SLIDER_RG_PREAMP, TBM_SETRANGEMIN,
          TRUE, 0);
      SendDlgItemMessage(hDlg, IDC_SLIDER_RG_PREAMP, TBM_SETRANGEMAX,
          TRUE, PREAMP_SCALE*(PREAMP_MAX-PREAMP_MIN));
      SendDlgItemMessage(hDlg, IDC_SLIDER_RG_PREAMP, TBM_SETPOS,
          TRUE, PREAMP_SCALE*(ini.rg.preamp-PREAMP_MIN));


      SendDlgItemMessage(hDlg, IDC_SLIDER_RG_ALGORITHM_PREAMP,
	      TBM_SETRANGEMIN, TRUE, 0);
      SendDlgItemMessage(hDlg, IDC_SLIDER_RG_ALGORITHM_PREAMP,
	      TBM_SETRANGEMAX, TRUE, PREAMP_SCALE*(PREAMP_MAX-PREAMP_MIN));
      SendDlgItemMessage(hDlg, IDC_SLIDER_RG_ALGORITHM_PREAMP, TBM_SETPOS,
          TRUE, PREAMP_SCALE*(ini.rg.algorithm_preamp-PREAMP_MIN));

      // sox
      CheckDlgButton(hDlg, IDC_CHECKBOX_SOX_LIMIT_SWITCH,
          ini.sox.limit_switch);

      j=i=0;

	  while (NULL!=limit_gain_prefs[i].label) {
        SendDlgItemMessage(hDlg, IDC_COMBO_SOX_LIMIT_GAIN, CB_ADDSTRING, 0,
            (LPARAM)limit_gain_prefs[i].label);

        if (0==strcmp(limit_gain_prefs[i].gain,ini.sox.limit))
		  j=i;

		++i;
	  }

      SendDlgItemMessage(hDlg, IDC_COMBO_SOX_LIMIT_GAIN, CB_SETCURSEL, j, 0);
      j=i=0;

	  while (NULL!=rg_algorithm_prefs[i].label) {
        SendDlgItemMessage(hDlg, IDC_COMBO_RG_ALGORITHM, CB_ADDSTRING, 0,
            (LPARAM)rg_algorithm_prefs[i].label);

        if (0==strcmp(rg_algorithm_prefs[i].algorithm,ini.rg.algorithm))
		  j=i;

		++i;
	  }

      SendDlgItemMessage(hDlg, IDC_COMBO_RG_ALGORITHM, CB_SETCURSEL, j, 0);
      SetDlgItemText(hDlg, IDC_STATIC_RG_ALGORITHM_PREAMP,
			  rg_algorithm_prefs[j].label2);
      SendDlgItemMessage(hDlg, IDC_COMBO_SOX_AMPLIFY, CB_ADDSTRING, 0,
          (LPARAM)GAIN);
      SendDlgItemMessage(hDlg, IDC_COMBO_SOX_AMPLIFY, CB_ADDSTRING, 0,
          (LPARAM)VOL);
      SendDlgItemMessage(hDlg, IDC_COMBO_SOX_AMPLIFY, CB_SETCURSEL,
          0==strcmp(GAIN,ini.sox.amplify)?0:1, 0);

      j=i=0;

	  while (NULL!=resample_prefs[i].label) {
        SendDlgItemMessage(hDlg, IDC_COMBO_SOX_RATE, CB_ADDSTRING, 0,
            (LPARAM)resample_prefs[i].label);

        if (resample_prefs[i].rate==ini.sox.sample_rate)
		  j=i;

		++i;
	  }

      SendDlgItemMessage(hDlg, IDC_COMBO_SOX_RATE, CB_SETCURSEL, j, 0);
      j=i=0;

	  while (NULL!=rate_last_prefs[i].label) {
        SendDlgItemMessage(hDlg, IDC_COMBO_SOX_RATE_LAST, CB_ADDSTRING, 0,
            (LPARAM)rate_last_prefs[i].label);

        if (rate_last_prefs[i].rate_last==ini.sox.rate_last)
		  j=i;

		++i;
	  }

      SendDlgItemMessage(hDlg, IDC_COMBO_SOX_RATE_LAST, CB_SETCURSEL, j, 0);
      j=i=0;

	  while (NULL!=quality_prefs[i].label) {
        SendDlgItemMessage(hDlg, IDC_COMBO_SOX_RATE_QUALITY, CB_ADDSTRING, 0,
            (LPARAM)quality_prefs[i].label);

        if (0==strcmp(quality_prefs[i].quality,ini.sox.quality))
		  j=i;

		++i;
	  }

      SendDlgItemMessage(hDlg, IDC_COMBO_SOX_RATE_QUALITY, CB_SETCURSEL, j, 0);

      j=i=0;

	  while (NULL!=phase_response_prefs[i].label) {
        SendDlgItemMessage(hDlg, IDC_COMBO_SOX_RATE_PHASE_RESPONSE, CB_ADDSTRING, 0,
            (LPARAM)phase_response_prefs[i].label);

        if (0==strcmp(phase_response_prefs[i].phase_response,ini.sox.phase_response))
		  j=i;

		++i;
	  }

      SendDlgItemMessage(hDlg, IDC_COMBO_SOX_RATE_PHASE_RESPONSE, CB_SETCURSEL, j, 0);
      CheckDlgButton(hDlg, IDC_CHECKBOX_SOX_LARGEST_MULTIPLE,
          ini.sox.largest_multiple);
      CheckDlgButton(hDlg, IDC_CHECKBOX_SOX_DITHER, ini.sox.dither);
      CheckDlgButton(hDlg, IDC_CHECKBOX_SOX_NOISE_SHAPING,
          ini.sox.noise_shaping);

      // plugin
      sprintf(buf, "%d", ini.plugin.timer_res);
      SetDlgItemText(hDlg, IDC_EDIT_PLUGIN_TIMER_RES, buf);
#ifdef __SOX_DECODER__
      CheckDlgButton(hDlg, IDC_CHECKBOX_PLUGIN_SOX_DECODER,
              ini.plugin.sox_decoder);
#endif // __SOX_DECODER__
      sprintf(buf, "%d", ini.plugin.drop_frames.before);
      CheckDlgButton(hDlg, IDC_CHECKBOX_PLUGIN_DROP_FRAMES,
              ini.plugin.drop_frames.on);
      sprintf(buf, "%d", ini.plugin.drop_frames.before);
      SetDlgItemText(hDlg, IDC_EDIT_PLUGIN_DROP_FRAMES_BEFORE, buf);
      sprintf(buf, "%d", ini.plugin.drop_frames.after);
      SetDlgItemText(hDlg, IDC_EDIT_PLUGIN_DROP_FRAMES_AFTER, buf);

      SendDlgItemMessage(hDlg, IDC_COMBO_PLUGIN_AC3_DECODER, CB_ADDSTRING, 0,
          (LPARAM)"libavcodec");
      SendDlgItemMessage(hDlg, IDC_COMBO_PLUGIN_AC3_DECODER, CB_ADDSTRING, 0,
          (LPARAM)"liba52");
      SendDlgItemMessage(hDlg, IDC_COMBO_PLUGIN_AC3_DECODER, CB_SETCURSEL,
          0==strcmp(_ac3_decoder,LIBAVCODEC)?0:1, 0);

#ifdef __CHOOSE_LIB__
      if (!a52_stub_valid)
        EnableWindow(GetDlgItem (hDlg, IDC_COMBO_PLUGIN_AC3_DECODER), FALSE);
#else
      EnableWindow(GetDlgItem (hDlg, IDC_COMBO_PLUGIN_AC3_DECODER), FALSE);
#endif

      SendDlgItemMessage(hDlg, IDC_COMBO_PLUGIN_MP3_DECODER, CB_ADDSTRING, 0,
          (LPARAM)"libavcodec");
      SendDlgItemMessage(hDlg, IDC_COMBO_PLUGIN_MP3_DECODER, CB_ADDSTRING, 0,
          (LPARAM)"libmad");
      SendDlgItemMessage(hDlg, IDC_COMBO_PLUGIN_MP3_DECODER, CB_SETCURSEL,
          0==strcmp(_mp3_decoder,LIBAVCODEC)?0:1, 0);

#ifdef __CHOOSE_LIB__
      if (!mad_stub_valid)
        EnableWindow(GetDlgItem (hDlg, IDC_COMBO_PLUGIN_MP3_DECODER), FALSE);
#else
      EnableWindow(GetDlgItem (hDlg, IDC_COMBO_PLUGIN_MP3_DECODER), FALSE);
#endif

      CheckDlgButton(hDlg, IDC_CHECKBOX_PLUGIN_FORCE_SEEK,
	      ini.plugin.force_seek);
      CheckDlgButton(hDlg, IDC_CHECKBOX_PLUGIN_PREFER_FLOAT,
	      ini.plugin.prefer_float);

      SendDlgItemMessage(hDlg,IDC_LIST_PLUGIN_TAG_PAGE,LB_RESETCONTENT,0,0);

	  /////////////////////
      j=i=0;

	  while (NULL!=pages[i].label) {
        SendDlgItemMessage(hDlg, IDC_LIST_PLUGIN_TAG_PAGE, LB_ADDSTRING, 0,
            (LPARAM)pages[i].label);

        if (pages[i].page==ini.plugin.tag_page)
		  j=i;

		++i;
	  }

      SendDlgItemMessage(hDlg, IDC_LIST_PLUGIN_TAG_PAGE, LB_SETCURSEL, j, 0);

	  /////////////////////
      j=i=0;

	  while (NULL!=pages[i].label) {
        SendDlgItemMessage(hDlg, IDC_LIST_PLUGIN_PATH_PAGE, LB_ADDSTRING, 0,
            (LPARAM)pages[i].label);

        if (pages[i].page==ini.plugin.path_page)
		  j=i;

		++i;
	  }

      SendDlgItemMessage(hDlg, IDC_LIST_PLUGIN_PATH_PAGE, LB_SETCURSEL, j, 0);

      // queues
      sprintf(buf, "%d", ini.queues.audio.min);
      SetDlgItemText(hDlg, IDC_EDIT_QUEUES_AUDIO_MIN, buf);
      sprintf(buf, "%d", ini.queues.audio.max);
      SetDlgItemText(hDlg, IDC_EDIT_QUEUES_AUDIO_MAX, buf);
      sprintf(buf, "%d", ini.queues.video.min);
      SetDlgItemText(hDlg, IDC_EDIT_QUEUES_VIDEO_MIN, buf);
      sprintf(buf, "%d", ini.queues.video.max);
      SetDlgItemText(hDlg, IDC_EDIT_QUEUES_VIDEO_MAX, buf);

      if (0==ini.sox.sample_rate)
        enable_resample(hDlg, FALSE);

      if (! ini.rg.mode)
        enable_rg(hDlg, FALSE);

      if (! ini.sox.dither)
        enable_dither(hDlg, FALSE);

      if (! ini.plugin.drop_frames.on)
        enable_drop_frames(hDlg, FALSE);

      enable_vol(hDlg, 0==strcmp(ini.sox.amplify,GAIN)?FALSE:TRUE);

      SetFocus(GetDlgItem (hDlg, IDC_EDIT_WINAMP_EXTS));
      break;
    case WM_NOTIFY:
      switch (LOWORD(wParam)) {
          case IDC_SLIDER_RG_PREAMP:
          sprintf(buf, "%.1f dB", 0.1*round(10.0*((double)SendDlgItemMessage(hDlg,
              IDC_SLIDER_RG_PREAMP, TBM_GETPOS, 0, 0)/PREAMP_SCALE+PREAMP_MIN)));
          SetDlgItemText(hDlg, IDC_STATIC_RG_PREAMP_DB, buf);
          break;

          case IDC_SLIDER_RG_ALGORITHM_PREAMP:
          sprintf(buf, "%.1f dB", 0.1*round(10.0*((double)SendDlgItemMessage(hDlg,
              IDC_SLIDER_RG_ALGORITHM_PREAMP, TBM_GETPOS, 0, 0)/PREAMP_SCALE+PREAMP_MIN)));
          SetDlgItemText(hDlg, IDC_STATIC_RG_ALGORITHM_PREAMP_DB, buf);
          break;
      }

      break;
    case WM_COMMAND:
      switch (LOWORD(wParam)) {
        case IDC_BUTTON_WINAMP_EXTS_DEF:
          SetDlgItemText(hDlg, IDC_EDIT_WINAMP_EXTS, DEF_EXTS);
          break;
        case IDC_COMBO_RG_ALGORITHM:
		  select_algorithm(hDlg, SendDlgItemMessage(hDlg,
              IDC_COMBO_RG_ALGORITHM, CB_GETCURSEL, 0, 0));
          break;
        case IDC_COMBO_RG_MODE:
          enable_rg(hDlg, 0<SendDlgItemMessage(hDlg,
              IDC_COMBO_RG_MODE, CB_GETCURSEL, 0, 0)?TRUE:FALSE);
          break;
        case IDC_COMBO_SOX_RATE:
          enable_resample(hDlg, 0<SendDlgItemMessage(hDlg,
              IDC_COMBO_SOX_RATE, CB_GETCURSEL, 0, 0)?TRUE:FALSE);
          break;
        case IDC_CHECKBOX_SOX_DITHER:
          enable_dither(hDlg, IsDlgButtonChecked (hDlg,
              LOWORD(wParam))?TRUE:FALSE);
          break;
        case IDC_CHECKBOX_PLUGIN_DROP_FRAMES:
          enable_drop_frames(hDlg, IsDlgButtonChecked (hDlg,
              LOWORD(wParam))?TRUE:FALSE);
          break;
        case IDC_COMBO_SOX_AMPLIFY:
        case IDC_CHECKBOX_SOX_LIMIT_SWITCH:
          enable_vol(hDlg, 1==SendDlgItemMessage(hDlg,
              IDC_COMBO_SOX_AMPLIFY, CB_GETCURSEL, 0, 0)?TRUE:FALSE);
          break;
        case IDC_BUTTON_QUEUES_AUDIO_DEF:
          sprintf(buf, "%d", PACKET_QUEUE_MIN);
          SetDlgItemText(hDlg, IDC_EDIT_QUEUES_AUDIO_MIN, buf);
          sprintf(buf, "%d", PACKET_QUEUE_MAX);
          SetDlgItemText(hDlg, IDC_EDIT_QUEUES_AUDIO_MAX, buf);
          break;
        case IDC_BUTTON_QUEUES_VIDEO_DEF:
          sprintf(buf, "%d", PACKET_QUEUE_MIN);
          SetDlgItemText(hDlg, IDC_EDIT_QUEUES_VIDEO_MIN, buf);
          sprintf(buf, "%d", PACKET_QUEUE_MAX);
          SetDlgItemText(hDlg, IDC_EDIT_QUEUES_VIDEO_MAX, buf);
          break;
        case IDOK:
          // winamp
          GetDlgItemText(hDlg,IDC_EDIT_WINAMP_EXTS,ini.winamp.exts,
              sizeof ini.winamp.exts);
          ini.winamp.bps=8*(2+SendDlgItemMessage(hDlg,
              IDC_COMBO_WINAMP_BPS, CB_GETCURSEL, 0, 0));
          ini.winamp.vis=IsDlgButtonChecked (hDlg,
              IDC_CHECKBOX_WINAMP_VIS)?1:0;
          ini.winamp.vid=IsDlgButtonChecked (hDlg,
              IDC_CHECKBOX_WINAMP_VID)?1:0;

          // rg
          ini.rg.mode=SendDlgItemMessage(hDlg, IDC_COMBO_RG_MODE,
              CB_GETCURSEL, 0, 0);
          tmp_rg_preamp=ini.rg.preamp
		      =0.1*round(10.0*((double)SendDlgItemMessage(hDlg,
              IDC_SLIDER_RG_PREAMP, TBM_GETPOS, 0, 0)/PREAMP_SCALE+PREAMP_MIN));
          tmp_rg_algorithm_preamp=ini.rg.algorithm_preamp
		      =0.1*round(10.0*((double)SendDlgItemMessage(hDlg,
              IDC_SLIDER_RG_ALGORITHM_PREAMP, TBM_GETPOS, 0, 0)
			  /PREAMP_SCALE+PREAMP_MIN));
	      strcpy(ini.rg.algorithm, rg_algorithm_prefs[SendDlgItemMessage(hDlg,
		      IDC_COMBO_RG_ALGORITHM, CB_GETCURSEL, 0, 0)].algorithm);
	      strcpy(tmp_rg_algorithm, ini.rg.algorithm);
          ini.rg.comment=IsDlgButtonChecked (hDlg,
              IDC_CHECKBOX_RG_COMMENT)?1:0;

          // sox
          ini.sox.limit_switch=IsDlgButtonChecked (hDlg,
              IDC_CHECKBOX_SOX_LIMIT_SWITCH)?1:0;
          strcpy(ini.sox.limit, limit_gain_prefs[SendDlgItemMessage(hDlg,
		      IDC_COMBO_SOX_LIMIT_GAIN, CB_GETCURSEL, 0, 0)].gain);
          ini.sox.sample_rate=resample_prefs[SendDlgItemMessage(hDlg,
		      IDC_COMBO_SOX_RATE, CB_GETCURSEL, 0, 0)].rate;
          ini.sox.rate_last=rate_last_prefs[SendDlgItemMessage(hDlg,
		      IDC_COMBO_SOX_RATE_LAST, CB_GETCURSEL, 0, 0)].rate_last;
          strcpy(ini.sox.quality, quality_prefs[SendDlgItemMessage(hDlg,
		      IDC_COMBO_SOX_RATE_QUALITY, CB_GETCURSEL, 0, 0)].quality);
          strcpy(ini.sox.phase_response, phase_response_prefs[SendDlgItemMessage(hDlg,
		      IDC_COMBO_SOX_RATE_PHASE_RESPONSE, CB_GETCURSEL, 0, 0)].phase_response);
          ini.sox.largest_multiple=IsDlgButtonChecked (hDlg,
              IDC_CHECKBOX_SOX_LARGEST_MULTIPLE)?1:0;
          ini.sox.dither=IsDlgButtonChecked (hDlg,
              IDC_CHECKBOX_SOX_DITHER)?1:0;
          ini.sox.noise_shaping=IsDlgButtonChecked (hDlg,
              IDC_CHECKBOX_SOX_NOISE_SHAPING)?1:0;
          strcpy(ini.sox.amplify,0==SendDlgItemMessage(hDlg,
              IDC_COMBO_SOX_AMPLIFY,CB_GETCURSEL,0,0)?GAIN:VOL);

          // plugin
          GetDlgItemText(hDlg,IDC_EDIT_PLUGIN_TIMER_RES, buf, sizeof buf);
          ini.plugin.timer_res=atoi(buf);
#ifdef __SOX_DECODER__
          ini.plugin.sox_decoder=IsDlgButtonChecked (hDlg,
              IDC_CHECKBOX_PLUGIN_SOX_DECODER)?1:0;
#endif // __SOX_DECODER__
          ini.plugin.drop_frames.on=IsDlgButtonChecked (hDlg,
              IDC_CHECKBOX_PLUGIN_DROP_FRAMES)?1:0;
          GetDlgItemText(hDlg,IDC_EDIT_PLUGIN_DROP_FRAMES_BEFORE, buf,
              sizeof buf);
          ini.plugin.drop_frames.before=atoi(buf);
          GetDlgItemText(hDlg,IDC_EDIT_PLUGIN_DROP_FRAMES_AFTER, buf,
              sizeof buf);
          ini.plugin.drop_frames.after=atoi(buf);
#ifdef __CHOOSE_LIB__
          strcpy(ini.plugin._ac3_decoder,0==SendDlgItemMessage(hDlg,
              IDC_COMBO_PLUGIN_AC3_DECODER, CB_GETCURSEL, 0, 0)
              ?LIBAVCODEC:LIBA52);
          strcpy(ini.plugin._mp3_decoder,0==SendDlgItemMessage(hDlg,
              IDC_COMBO_PLUGIN_MP3_DECODER, CB_GETCURSEL, 0, 0)
              ?LIBAVCODEC:LIBMAD);
#endif
          ini.plugin.force_seek=IsDlgButtonChecked (hDlg,
              IDC_CHECKBOX_PLUGIN_FORCE_SEEK)?1:0;
          ini.plugin.prefer_float=IsDlgButtonChecked (hDlg,
              IDC_CHECKBOX_PLUGIN_PREFER_FLOAT)?1:0;
          ini.plugin.tag_page=pages[SendDlgItemMessage(hDlg,
		      IDC_LIST_PLUGIN_TAG_PAGE, LB_GETCURSEL, 0, 0)].page;
          ini.plugin.path_page=pages[SendDlgItemMessage(hDlg,
		      IDC_LIST_PLUGIN_PATH_PAGE, LB_GETCURSEL, 0, 0)].page;

          // queues
          GetDlgItemText(hDlg,IDC_EDIT_QUEUES_AUDIO_MIN, buf, sizeof buf);
          ini.queues.audio.min=atoi(buf);
          GetDlgItemText(hDlg,IDC_EDIT_QUEUES_AUDIO_MAX, buf, sizeof buf);
          ini.queues.audio.max=atoi(buf);
          GetDlgItemText(hDlg,IDC_EDIT_QUEUES_VIDEO_MIN, buf, sizeof buf);
          ini.queues.video.min=atoi(buf);
          GetDlgItemText(hDlg,IDC_EDIT_QUEUES_VIDEO_MAX, buf, sizeof buf);
          ini.queues.video.max=atoi(buf);

          save_ini();
          get_exts();
          EndDialog (hDlg, 1);
          return TRUE;
        case IDCANCEL:
          EndDialog (hDlg, 0);
          return TRUE;
      }

      break;
  }

  return FALSE;
}

static char *get_gain_val(struct tag *tag)
{
  return
      1==ini.rg.mode?tag->replaygain_track_gain:
      2==ini.rg.mode?tag->replaygain_album_gain:
	  NULL;
}

static char *get_peak_val(struct tag *tag)
{
  return
      1==ini.rg.mode?tag->replaygain_track_peak:
      2==ini.rg.mode?tag->replaygain_album_peak:
	  NULL;
}

static int ends_with(char *path, char *sfx)
{
  int l1=strlen(path);
  int l2=strlen(sfx);

  return l2<=l1&&0==stricmp(path+l1-l2,sfx);
}

// dB
void prefs_set_gain(struct tag *tag)
{
  char *gain_val=get_gain_val(tag);
  char *peak_val=get_peak_val(tag);
  char *interpreted=NULL;

  if (NULL==gain_val||0==*gain_val) {
    tag->gain=0.0;
    tag->q=0.0;
  }
  else if (0==stricmp("EBU R128", tag->replaygain_algorithm)
      ||ends_with(gain_val, "LU")) {

    if (0==strcmp("r128",ini.rg.algorithm))
      tag->gain=ini.rg.preamp+atof(gain_val);
	else
      tag->gain=ini.rg.algorithm_preamp+ini.rg.preamp+atof(gain_val);

    tag->q=pow(2.0,tag->gain/6.0);
    interpreted="EBU R128";
#ifdef __DEBUG__
fprintf(stderr, "EBU R128: %.1f dB\n", tag->gain);
#endif // __DEBUG__
  }
  else {
    if (0==strcmp("rg",ini.rg.algorithm))
      tag->gain=ini.rg.preamp+atof(gain_val);
	else
      tag->gain=ini.rg.algorithm_preamp+ini.rg.preamp+atof(gain_val);

    tag->q=pow(2.0,tag->gain/6.0);
    interpreted="ReplayGain";
#ifdef __DEBUG__
fprintf(stderr, "ReplayGain: %.1f dB\n", tag->gain);
#endif // __DEBUG__
  }

  if (NULL!=interpreted&&ini.rg.comment) {
    if (NULL!=peak_val&&0!=*peak_val) {
      double peak=20*log10(atof(peak_val));
      double new_peak=peak+tag->gain;

	  sprintf(tag->comment,
		  "%s - "
	      "Original Peak: %.1f dBFS"
	      ", Gain: %.1f dB"
	      ", Resulting in %s: %.1f dBFS",
		  interpreted,
		  peak,
		  tag->gain,
		  new_peak<0.0?"Headroom":"Clipping",
		  new_peak);
    }
	else {
	  sprintf(tag->comment,
		  "%s - "
	      "Gain: %.1f dB",
		  interpreted,
		  tag->gain);
	}
  }
}
