/*
 * samplebuf.c
 * Copyright (C) 2010 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <stdio.h>
#include "plugin.h"

samplebuf_t *samplebuf_open(samplebuf_t *buf, plugin_t *plugin,
    track_t *track, const char *fn, int nch, int rate, int bps,
	int length)
{
#ifdef __XDEBUG__
  fprintf(stderr, "samplebuf_open()\n");
#endif // __XDEBUG__
  memset(buf, 0, sizeof *buf);

  buf->plugin=plugin;
  buf->track=track;
  _snprintf(buf->fn, sizeof buf->fn, "%s", fn); 

  buf->nsam=576;
  //buf->nch=nch;
  //buf->nch=2;
  buf->nch=1<nch?2:1;

  if (0==ini.sox.sample_rate)
    buf->rate=rate;
  else if (0==ini.sox.largest_multiple)
    buf->rate=ini.sox.sample_rate;
  else if (rate<ini.sox.sample_rate)
    buf->rate=(ini.sox.sample_rate/rate)*rate;
  else
    buf->rate=rate;

  buf->bps=ini.winamp.bps;
  buf->bytes=buf->bps/8;
  buf->length=length;

  return buf;
}

samplebuf_t *samplebuf_start(samplebuf_t *buf)
{
#ifdef __XDEBUG__
  fprintf(stderr, "samplebuf_start()\n");
#endif // __XDEBUG__
  buf->size=buf->nsam*buf->nch*buf->bytes;
  buf->samples=malloc(buf->size*2); 
  buf->mp=buf->samples+buf->size;
  buf->wp=buf->samples;
  buf->state=SAMPLEBUF_START;
#ifdef __XDEBUG__
  fprintf(stderr, "%d: %p, %p, %p\n",buf->size,buf->samples,buf->wp,buf->mp);
#endif // __XDEBUG__

  buf->hEventRead=CreateEvent(
    NULL,           // lpEventAttributes
    FALSE,          // bManualReset
    FALSE,          // bInitialState
    NULL            // lpName
  );

  buf->hEventWrite=CreateEvent(
    NULL,           // lpEventAttributes
    FALSE,          // bManualReset
    FALSE,          // bInitialState
    NULL            // lpName
  );

  return buf;
}

void samplebuf_close(samplebuf_t *buf)
{
#ifdef __XDEBUG__
  fprintf(stderr, "samplebuf_close()\n");
#endif // __XDEBUG__
}

void samplebuf_stop(samplebuf_t *buf)
{
#ifdef __XDEBUG__
  fprintf(stderr, "samplebuf_stop()\n");
#endif // __XDEBUG__
  if (NULL!=buf->hEventWrite) {
    CloseHandle(buf->hEventWrite);
    buf->hEventWrite=NULL;
  }

  if (NULL!=buf->hEventRead) {
    CloseHandle(buf->hEventRead);
    buf->hEventRead=NULL;
  }

  if (NULL!=buf->samples) {
    free(buf->samples);
    buf->samples=NULL;
  }
}

void samplebuf_reset(samplebuf_t *buf)
{
#ifdef __XDEBUG__
  fprintf(stderr, "samplebuf_reset()\n");
#endif // __XDEBUG__
  while (samplebuf_read(buf))
    ;

  buf->state=SAMPLEBUF_START;
}

void samplebuf_signal_done(samplebuf_t *buf)
{
#ifdef __XDEBUG__
  fprintf(stderr, "samplebuf_signal_done()\n");
#endif // __XDEBUG__
  buf->state=SAMPLEBUF_DONE;
  SetEvent(buf->hEventRead);
  SetEvent(buf->hEventWrite);
}

void samplebuf_signal_read(samplebuf_t *buf)
{
#ifdef __XDEBUG__
  fprintf(stderr, "samplebuf_signal_read()\n");
#endif // __XDEBUG__
  buf->state=SAMPLEBUF_READ;
  SetEvent(buf->hEventRead);
}

void samplebuf_signal_write(samplebuf_t *buf)
{
#ifdef __XDEBUG__
  fprintf(stderr, "samplebuf_signal_write()\n");
#endif // __XDEBUG__
  buf->state=SAMPLEBUF_WRITE;
  SetEvent(buf->hEventWrite);
}

int samplebuf_wait_read(samplebuf_t *buf)
{
#ifdef __XDEBUG__
  fprintf(stderr, "samplebuf_wait_read()\n");
#endif // __XDEBUG__

  while (SAMPLEBUF_DONE!=buf->state&&SAMPLEBUF_READ!=buf->state)
    WaitForSingleObject(buf->hEventRead, INFINITE);

  return buf->state;
}

int samplebuf_wait_write(samplebuf_t *buf)
{
#ifdef __XDEBUG__
  fprintf(stderr, "samplebuf_wait_write()\n");
#endif // __XDEBUG__

  while (SAMPLEBUF_DONE!=buf->state&&SAMPLEBUF_WRITE!=buf->state)
    WaitForSingleObject(buf->hEventWrite, INFINITE);

  return buf->state;
}

int samplebuf_read(samplebuf_t *buf)
{
#ifdef __XDEBUG__
  fprintf(stderr, "samplebuf_read()\n");
#endif // __XDEBUG__
	buf->wp=buf->samples;

	if (SAMPLEBUF_DONE!=buf->state) {
	  samplebuf_signal_write(buf);
	  samplebuf_wait_read(buf);
    }

	return buf->wp-buf->samples;
}
